@@ -1,7 +1,7 @@
 /*
  *  be_package.c : backend for packages
  *
- *  Copyright (c) 2006-2011 Pacman Development Team <pacman-dev@archlinux.org>
+ *  Copyright (c) 2006-2013 Pacman Development Team <pacman-dev@archlinux.org>
  *  Copyright (c) 2002-2006 by Judd Vinet <jvinet@zeroflux.org>
  *
  *  This program is free software; you can redistribute it and/or modify
@@ -18,11 +18,12 @@
  *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
  */
 
-#include "config.h"
-
 #include <stdlib.h>
 #include <string.h>
 #include <errno.h>
+#include <sys/types.h>
+#include <sys/stat.h>
+#include <fcntl.h>
 
 /* libarchive */
 #include <archive.h>
@@ -31,11 +32,18 @@
 /* libalpm */
 #include "alpm_list.h"
 #include "alpm.h"
+#include "libarchive-compat.h"
 #include "util.h"
 #include "log.h"
 #include "handle.h"
 #include "package.h"
-#include "deps.h" /* _alpm_splitdep */
+#include "deps.h"
+#include "filelist.h"
+
+struct package_changelog {
+	struct archive *archive;
+	int fd;
+};
 
 /**
  * Open a package changelog for reading. Similar to fopen in functionality,
@@ -47,31 +55,38 @@ static void *_package_changelog_open(alpm_pkg_t *pkg)
 {
 	ASSERT(pkg != NULL, return NULL);
 
-	struct archive *archive = NULL;
+	struct package_changelog *changelog;
+	struct archive *archive;
 	struct archive_entry *entry;
 	const char *pkgfile = pkg->origin_data.file;
+	struct stat buf;
+	int fd;
 
-	if((archive = archive_read_new()) == NULL) {
-		RET_ERR(pkg->handle, ALPM_ERR_LIBARCHIVE, NULL);
-	}
-
-	archive_read_support_compression_all(archive);
-	archive_read_support_format_all(archive);
-
-	if(archive_read_open_filename(archive, pkgfile,
-				ALPM_BUFFER_SIZE) != ARCHIVE_OK) {
-		RET_ERR(pkg->handle, ALPM_ERR_PKG_OPEN, NULL);
+	fd = _alpm_open_archive(pkg->handle, pkgfile, &buf,
+			&archive, ALPM_ERR_PKG_OPEN);
+	if(fd < 0) {
+		return NULL;
 	}
 
 	while(archive_read_next_header(archive, &entry) == ARCHIVE_OK) {
 		const char *entry_name = archive_entry_pathname(entry);
 
 		if(strcmp(entry_name, ".CHANGELOG") == 0) {
-			return archive;
+			changelog = malloc(sizeof(struct package_changelog));
+			if(!changelog) {
+				pkg->handle->pm_errno = ALPM_ERR_MEMORY;
+				_alpm_archive_read_free(archive);
+				CLOSE(fd);
+				return NULL;
+			}
+			changelog->archive = archive;
+			changelog->fd = fd;
+			return changelog;
 		}
 	}
 	/* we didn't find a changelog */
-	archive_read_finish(archive);
+	_alpm_archive_read_free(archive);
+	CLOSE(fd);
 	errno = ENOENT;
 
 	return NULL;
@@ -89,7 +104,8 @@ static void *_package_changelog_open(alpm_pkg_t *pkg)
 static size_t _package_changelog_read(void *ptr, size_t size,
 		const alpm_pkg_t UNUSED *pkg, void *fp)
 {
-	ssize_t sret = archive_read_data((struct archive *)fp, ptr, size);
+	struct package_changelog *changelog = fp;
+	ssize_t sret = archive_read_data(changelog->archive, ptr, size);
 	/* Report error (negative values) */
 	if(sret < 0) {
 		RET_ERR(pkg->handle, ALPM_ERR_LIBARCHIVE, 0);
@@ -107,7 +123,12 @@ static size_t _package_changelog_read(void *ptr, size_t size,
  */
 static int _package_changelog_close(const alpm_pkg_t UNUSED *pkg, void *fp)
 {
-	return archive_read_finish((struct archive *)fp);
+	int ret;
+	struct package_changelog *changelog = fp;
+	ret = _alpm_archive_read_free(changelog->archive);
+	CLOSE(changelog->fd);
+	free(changelog);
+	return ret;
 }
 
 /** Package file operations struct accessor. We implement this as a method
@@ -149,21 +170,24 @@ static int parse_descfile(alpm_handle_t *handle, struct archive *a, alpm_pkg_t *
 
 	/* loop until we reach EOF or other error */
 	while((ret = _alpm_archive_fgets(a, &buf)) == ARCHIVE_OK) {
-		size_t len = _alpm_strip_newline(buf.line);
+		size_t len = _alpm_strip_newline(buf.line, buf.real_line_size);
 
 		linenum++;
-		if(len == 0 || buf.line[0] == '#') {
+		key = buf.line;
+		if(len == 0 || key[0] == '#') {
 			continue;
 		}
-		ptr = buf.line;
-		key = strsep(&ptr, "=");
-		if(key == NULL || ptr == NULL) {
-			_alpm_log(handle, ALPM_LOG_DEBUG, "%s: syntax error in description file line %d\n",
-								newpkg->name ? newpkg->name : "error", linenum);
+		/* line is always in this format: "key = value"
+		 * we can be sure the " = " exists, so look for that */
+		ptr = memchr(key, ' ', len);
+		if(!ptr || (size_t)(ptr - key + 2) > len || memcmp(ptr, " = ", 3) != 0) {
+			_alpm_log(handle, ALPM_LOG_DEBUG,
+					"%s: syntax error in description file line %d\n",
+					newpkg->name ? newpkg->name : "error", linenum);
 		} else {
-			key = _alpm_strtrim(key);
-			while(*ptr == ' ') ptr++;
-			ptr = _alpm_strtrim(ptr);
+			/* NULL the end of the key portion, move ptr to start of value */
+			*ptr = '\0';
+			ptr += 3;
 			if(strcmp(key, "pkgname") == 0) {
 				STRDUP(newpkg->name, ptr, return -1);
 				newpkg->name_hash = _alpm_hash_sdbm(newpkg->name);
@@ -192,7 +216,8 @@ static int parse_descfile(alpm_handle_t *handle, struct archive *a, alpm_pkg_t *
 				alpm_depend_t *dep = _alpm_splitdep(ptr);
 				newpkg->depends = alpm_list_add(newpkg->depends, dep);
 			} else if(strcmp(key, "optdepend") == 0) {
-				newpkg->optdepends = alpm_list_add(newpkg->optdepends, strdup(ptr));
+				alpm_depend_t *optdep = _alpm_splitdep(ptr);
+				newpkg->optdepends = alpm_list_add(newpkg->optdepends, optdep);
 			} else if(strcmp(key, "conflict") == 0) {
 				alpm_depend_t *conflict = _alpm_splitdep(ptr);
 				newpkg->conflicts = alpm_list_add(newpkg->conflicts, conflict);
@@ -225,53 +250,6 @@ static int parse_descfile(alpm_handle_t *handle, struct archive *a, alpm_pkg_t *
 	return 0;
 }
 
-static void files_merge(alpm_file_t a[], alpm_file_t b[], alpm_file_t c[],
-		size_t m, size_t n)
-{
-	size_t i = 0, j = 0, k = 0;
-	while(i < m && j < n) {
-		if(strcmp(a[i].name, b[j].name) < 0) {
-			c[k++] = a[i++];
-		} else {
-			c[k++] = b[j++];
-		}
-	}
-	while(i < m) {
-		c[k++] = a[i++];
-	}
-	while(j < n) {
-		c[k++] = b[j++];
-	}
-}
-
-static alpm_file_t *files_msort(alpm_file_t *files, size_t n)
-{
-	alpm_file_t *work;
-	size_t blocksize = 1;
-
-	CALLOC(work, n, sizeof(alpm_file_t), return NULL);
-
-	for(blocksize = 1; blocksize < n; blocksize *= 2) {
-		size_t i, max_extent = 0;
-		for(i = 0; i < n - blocksize; i += 2 * blocksize) {
-			/* this limits our actual merge to the length of the array, since we will
-			 * not likely be a perfect power of two. */
-			size_t right_blocksize = blocksize;
-			if(i + blocksize * 2 > n) {
-				right_blocksize = n - i - blocksize;
-			}
-			files_merge(files + i, files + i + blocksize, work + i,
-					blocksize, right_blocksize);
-			max_extent = i + blocksize + right_blocksize;
-		}
-		/* ensure we only copy what we actually touched on this merge pass,
-		 * no more, no less */
-		memcpy(files, work, max_extent * sizeof(alpm_file_t));
-	}
-	free(work);
-	return files;
-}
-
 /**
  * Validate a package.
  * @param handle the context handle
@@ -280,11 +258,12 @@ static alpm_file_t *files_msort(alpm_file_t *files, size_t n)
  * sha256sum, and/or base64 signature)
  * @param level the required level of signature verification
  * @param sigdata signature data from the package to pass back
+ * @param validation successful validations performed on the package file
  * @return 0 if package is fully valid, -1 and pm_errno otherwise
  */
 int _alpm_pkg_validate_internal(alpm_handle_t *handle,
 		const char *pkgfile, alpm_pkg_t *syncpkg, alpm_siglevel_t level,
-		alpm_siglist_t **sigdata)
+		alpm_siglist_t **sigdata, alpm_pkgvalidation_t *validation)
 {
 	int has_sig;
 	handle->pm_errno = 0;
@@ -294,8 +273,15 @@ int _alpm_pkg_validate_internal(alpm_handle_t *handle,
 	}
 
 	/* attempt to access the package file, ensure it exists */
-	if(access(pkgfile, R_OK) != 0) {
-		RET_ERR(handle, ALPM_ERR_PKG_NOT_FOUND, -1);
+	if(_alpm_access(handle, NULL, pkgfile, R_OK) != 0) {
+		if(errno == ENOENT) {
+			handle->pm_errno = ALPM_ERR_PKG_NOT_FOUND;
+		} else if(errno == EACCES) {
+			handle->pm_errno = ALPM_ERR_BADPERMS;
+		} else {
+			handle->pm_errno = ALPM_ERR_PKG_OPEN;
+		}
+		return -1;
 	}
 
 	/* can we get away with skipping checksums? */
@@ -316,17 +302,23 @@ int _alpm_pkg_validate_internal(alpm_handle_t *handle,
 		if(syncpkg->md5sum && !syncpkg->sha256sum) {
 			_alpm_log(handle, ALPM_LOG_DEBUG, "md5sum: %s\n", syncpkg->md5sum);
 			_alpm_log(handle, ALPM_LOG_DEBUG, "checking md5sum for %s\n", pkgfile);
-			if(_alpm_test_checksum(pkgfile, syncpkg->md5sum, ALPM_CSUM_MD5) != 0) {
+			if(_alpm_test_checksum(pkgfile, syncpkg->md5sum, ALPM_PKG_VALIDATION_MD5SUM) != 0) {
 				RET_ERR(handle, ALPM_ERR_PKG_INVALID_CHECKSUM, -1);
 			}
+			if(validation) {
+				*validation |= ALPM_PKG_VALIDATION_MD5SUM;
+			}
 		}
 
 		if(syncpkg->sha256sum) {
 			_alpm_log(handle, ALPM_LOG_DEBUG, "sha256sum: %s\n", syncpkg->sha256sum);
 			_alpm_log(handle, ALPM_LOG_DEBUG, "checking sha256sum for %s\n", pkgfile);
-			if(_alpm_test_checksum(pkgfile, syncpkg->sha256sum, ALPM_CSUM_SHA256) != 0) {
+			if(_alpm_test_checksum(pkgfile, syncpkg->sha256sum, ALPM_PKG_VALIDATION_SHA256SUM) != 0) {
 				RET_ERR(handle, ALPM_ERR_PKG_INVALID_CHECKSUM, -1);
 			}
+			if(validation) {
+				*validation |= ALPM_PKG_VALIDATION_SHA256SUM;
+			}
 		}
 	}
 
@@ -340,6 +332,13 @@ int _alpm_pkg_validate_internal(alpm_handle_t *handle,
 			handle->pm_errno = ALPM_ERR_PKG_INVALID_SIG;
 			return -1;
 		}
+		if(validation && has_sig) {
+			*validation |= ALPM_PKG_VALIDATION_SIGNATURE;
+		}
+	}
+
+	if(validation && !*validation) {
+		*validation = ALPM_PKG_VALIDATION_NONE;
 	}
 
 	return 0;
@@ -355,10 +354,10 @@ int _alpm_pkg_validate_internal(alpm_handle_t *handle,
 alpm_pkg_t *_alpm_pkg_load_internal(alpm_handle_t *handle,
 		const char *pkgfile, int full)
 {
-	int ret, config = 0;
+	int ret, fd, config = 0;
 	struct archive *archive;
 	struct archive_entry *entry;
-	alpm_pkg_t *newpkg = NULL;
+	alpm_pkg_t *newpkg;
 	struct stat st;
 	size_t files_size = 0;
 
@@ -366,33 +365,26 @@ alpm_pkg_t *_alpm_pkg_load_internal(alpm_handle_t *handle,
 		RET_ERR(handle, ALPM_ERR_WRONG_ARGS, NULL);
 	}
 
-	/* attempt to stat the package file, ensure it exists */
-	if(stat(pkgfile, &st) == 0) {
-		newpkg = _alpm_pkg_new();
-		if(newpkg == NULL) {
-			RET_ERR(handle, ALPM_ERR_MEMORY, NULL);
+	fd = _alpm_open_archive(handle, pkgfile, &st, &archive, ALPM_ERR_PKG_OPEN);
+	if(fd < 0) {
+		if(errno == ENOENT) {
+			handle->pm_errno = ALPM_ERR_PKG_NOT_FOUND;
+		} else if(errno == EACCES) {
+			handle->pm_errno = ALPM_ERR_BADPERMS;
+		} else {
+			handle->pm_errno = ALPM_ERR_PKG_OPEN;
 		}
-		newpkg->filename = strdup(pkgfile);
-		newpkg->size = st.st_size;
-	} else {
-		/* couldn't stat the pkgfile, return an error */
-		RET_ERR(handle, ALPM_ERR_PKG_NOT_FOUND, NULL);
+		return NULL;
 	}
 
-	/* try to create an archive object to read in the package */
-	if((archive = archive_read_new()) == NULL) {
-		_alpm_pkg_free(newpkg);
-		RET_ERR(handle, ALPM_ERR_LIBARCHIVE, NULL);
-	}
-
-	archive_read_support_compression_all(archive);
-	archive_read_support_format_all(archive);
-
-	if(archive_read_open_filename(archive, pkgfile,
-				ALPM_BUFFER_SIZE) != ARCHIVE_OK) {
-		handle->pm_errno = ALPM_ERR_PKG_OPEN;
+	newpkg = _alpm_pkg_new();
+	if(newpkg == NULL) {
+		handle->pm_errno = ALPM_ERR_MEMORY;
 		goto error;
 	}
+	STRDUP(newpkg->filename, pkgfile,
+			handle->pm_errno = ALPM_ERR_MEMORY; goto error);
+	newpkg->size = st.st_size;
 
 	_alpm_log(handle, ALPM_LOG_DEBUG, "starting package load for %s\n", pkgfile);
 
@@ -439,7 +431,7 @@ alpm_pkg_t *_alpm_pkg_load_internal(alpm_handle_t *handle,
 				newfiles = realloc(newpkg->files.files,
 						sizeof(alpm_file_t) * files_size);
 				if(!newfiles) {
-					ALLOC_FAIL(sizeof(alpm_file_t) * files_size);
+					_alpm_alloc_fail(sizeof(alpm_file_t) * files_size);
 					goto error;
 				}
 				/* ensure all new memory is zeroed out, in both the initial
@@ -480,14 +472,16 @@ alpm_pkg_t *_alpm_pkg_load_internal(alpm_handle_t *handle,
 		goto pkg_invalid;
 	}
 
-	archive_read_finish(archive);
+	_alpm_archive_read_free(archive);
+	CLOSE(fd);
 
 	/* internal fields for package struct */
-	newpkg->origin = PKG_FROM_FILE;
+	newpkg->origin = ALPM_PKG_FROM_FILE;
 	newpkg->origin_data.file = strdup(pkgfile);
 	newpkg->ops = get_file_pkg_ops();
 	newpkg->handle = handle;
 	newpkg->infolevel = INFRQ_BASE | INFRQ_DESC | INFRQ_SCRIPTLET;
+	newpkg->validation = ALPM_PKG_VALIDATION_NONE;
 
 	if(full) {
 		if(newpkg->files.files) {
@@ -497,8 +491,9 @@ alpm_pkg_t *_alpm_pkg_load_internal(alpm_handle_t *handle,
 			/* "checking for conflicts" requires a sorted list, ensure that here */
 			_alpm_log(handle, ALPM_LOG_DEBUG,
 					"sorting package filelist for %s\n", pkgfile);
-			newpkg->files.files = files_msort(newpkg->files.files,
-					newpkg->files.count);
+
+			qsort(newpkg->files.files, newpkg->files.count,
+					sizeof(alpm_file_t), _alpm_files_cmp);
 		}
 		newpkg->infolevel |= INFRQ_FILES;
 	}
@@ -509,18 +504,89 @@ pkg_invalid:
 	handle->pm_errno = ALPM_ERR_PKG_INVALID;
 error:
 	_alpm_pkg_free(newpkg);
-	archive_read_finish(archive);
+	_alpm_archive_read_free(archive);
+	if(fd >= 0) {
+		CLOSE(fd);
+	}
 
 	return NULL;
 }
 
+static int read_sigfile(const char *sigpath, unsigned char **sig)
+{
+	struct stat st;
+	FILE *fp;
+
+	if(stat(sigpath, &st) != 0) {
+		return -1;
+	}
+
+	MALLOC(*sig, st.st_size, return -1);
+
+	if((fp = fopen(sigpath, "rb")) == NULL) {
+		free(*sig);
+		return -1;
+	}
+
+	if(fread(*sig, st.st_size, 1, fp) != 1) {
+		free(*sig);
+		fclose(fp);
+		return -1;
+	}
+
+	fclose(fp);
+	return st.st_size;
+}
+
 int SYMEXPORT alpm_pkg_load(alpm_handle_t *handle, const char *filename, int full,
 		alpm_siglevel_t level, alpm_pkg_t **pkg)
 {
+	alpm_pkgvalidation_t validation = 0;
+	char *sigpath;
+
 	CHECK_HANDLE(handle, return -1);
 	ASSERT(pkg != NULL, RET_ERR(handle, ALPM_ERR_WRONG_ARGS, -1));
 
-	if(_alpm_pkg_validate_internal(handle, filename, NULL, level, NULL) == -1) {
+	sigpath = _alpm_sigpath(handle, filename);
+	if(sigpath && !_alpm_access(handle, NULL, sigpath, R_OK)) {
+		if(level & ALPM_SIG_PACKAGE) {
+			alpm_list_t *keys = NULL;
+			int fail = 0;
+			unsigned char *sig = NULL;
+			int len = read_sigfile(sigpath, &sig);
+
+			if(len == -1) {
+				_alpm_log(handle, ALPM_LOG_ERROR,
+					_("failed to read signature file: %s\n"), sigpath);
+				free(sigpath);
+				return -1;
+			}
+
+			if(_alpm_extract_keyid(handle, filename, sig, len, &keys) == 0) {
+				alpm_list_t *k;
+				for(k = keys; k; k = k->next) {
+					char *key = k->data;
+					if(_alpm_key_in_keychain(handle, key) == 0) {
+						if(_alpm_key_import(handle, key) == -1) {
+							fail = 1;
+						}
+					}
+				}
+				FREELIST(keys);
+			}
+
+			free(sig);
+
+			if(fail) {
+				_alpm_log(handle, ALPM_LOG_ERROR, _("required key missing from keyring\n"));
+				return -1;
+			}
+		}
+	}
+	free(sigpath);
+
+	if(_alpm_pkg_validate_internal(handle, filename, NULL, level, NULL,
+				&validation) == -1) {
 		/* pm_errno is set by pkg_validate */
 		return -1;
 	}
@@ -529,6 +595,7 @@ int SYMEXPORT alpm_pkg_load(alpm_handle_t *handle, const char *filename, int ful
 		/* pm_errno is set by pkg_load */
 		return -1;
 	}
+	(*pkg)->validation = validation;
 
 	return 0;
 }
