@@ -1,7 +1,7 @@
 /*
  *  conflict.c
  *
- *  Copyright (c) 2006-2011 Pacman Development Team <pacman-dev@archlinux.org>
+ *  Copyright (c) 2006-2013 Pacman Development Team <pacman-dev@archlinux.org>
  *  Copyright (c) 2002-2006 by Judd Vinet <jvinet@zeroflux.org>
  *  Copyright (c) 2005 by Aurelien Foret <orelien@chez.com>
  *  Copyright (c) 2006 by David Kimpe <dnaku@frugalware.org>
@@ -22,8 +22,6 @@
  *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
  */
 
-#include "config.h"
-
 #include <stdlib.h>
 #include <stdio.h>
 #include <string.h>
@@ -40,7 +38,11 @@
 #include "util.h"
 #include "log.h"
 #include "deps.h"
+#include "filelist.h"
 
+/**
+ * @brief Creates a new conflict.
+ */
 static alpm_conflict_t *conflict_new(alpm_pkg_t *pkg1, alpm_pkg_t *pkg2,
 		alpm_depend_t *reason)
 {
@@ -57,6 +59,9 @@ static alpm_conflict_t *conflict_new(alpm_pkg_t *pkg1, alpm_pkg_t *pkg2,
 	return conflict;
 }
 
+/**
+ * @brief Free a conflict and its members.
+ */
 void _alpm_conflict_free(alpm_conflict_t *conflict)
 {
 	FREE(conflict->package2);
@@ -64,6 +69,9 @@ void _alpm_conflict_free(alpm_conflict_t *conflict)
 	FREE(conflict);
 }
 
+/**
+ * @brief Creates a copy of a conflict.
+ */
 alpm_conflict_t *_alpm_conflict_dup(const alpm_conflict_t *conflict)
 {
 	alpm_conflict_t *newconflict;
@@ -78,6 +86,14 @@ alpm_conflict_t *_alpm_conflict_dup(const alpm_conflict_t *conflict)
 	return newconflict;
 }
 
+/**
+ * @brief Searches for a conflict in a list.
+ *
+ * @param needle conflict to search for
+ * @param haystack list of conflicts to search
+ *
+ * @return 1 if needle is in haystack, 0 otherwise
+ */
 static int conflict_isin(alpm_conflict_t *needle, alpm_list_t *haystack)
 {
 	alpm_list_t *i;
@@ -94,12 +110,16 @@ static int conflict_isin(alpm_conflict_t *needle, alpm_list_t *haystack)
 	return 0;
 }
 
-/** Adds the pkg1/pkg2 conflict to the baddeps list.
+/**
+ * @brief Adds the pkg1/pkg2 conflict to the baddeps list.
+ *
  * @param handle the context handle
  * @param baddeps list to add conflict to
  * @param pkg1 first package
  * @param pkg2 package causing conflict
  * @param reason reason for this conflict
+ *
+ * @return 0 on success, -1 on error
  */
 static int add_conflict(alpm_handle_t *handle, alpm_list_t **baddeps,
 		alpm_pkg_t *pkg1, alpm_pkg_t *pkg2, alpm_depend_t *reason)
@@ -120,11 +140,13 @@ static int add_conflict(alpm_handle_t *handle, alpm_list_t **baddeps,
 	return 0;
 }
 
-/** Check if packages from list1 conflict with packages from list2.
- * This looks at the conflicts fields of all packages from list1, and sees
- * if they match packages from list2.
- * If a conflict (pkg1, pkg2) is found, it is added to the baddeps list
- * in this order if order >= 0, or reverse order (pkg2,pkg1) otherwise.
+/**
+ * @brief Check if packages from list1 conflict with packages from list2.
+ *
+ * @details This looks at the conflicts fields of all packages from list1, and
+ * sees if they match packages from list2. If a conflict (pkg1, pkg2) is found,
+ * it is added to the baddeps list in this order if order >= 0, or reverse
+ * order (pkg2,pkg1) otherwise.
  *
  * @param handle the context handle
  * @param list1 first list of packages
@@ -134,7 +156,8 @@ static int add_conflict(alpm_handle_t *handle, alpm_list_t **baddeps,
  */
 static void check_conflict(alpm_handle_t *handle,
 		alpm_list_t *list1, alpm_list_t *list2,
-		alpm_list_t **baddeps, int order) {
+		alpm_list_t **baddeps, int order)
+{
 	alpm_list_t *i;
 
 	if(!baddeps) {
@@ -169,7 +192,14 @@ static void check_conflict(alpm_handle_t *handle,
 	}
 }
 
-/* Check for inter-conflicts */
+/**
+ * @brief Check for inter-conflicts in a list of packages.
+ *
+ * @param handle the context handle
+ * @param packages list of packages to check
+ *
+ * @return list of conflicts
+ */
 alpm_list_t *_alpm_innerconflicts(alpm_handle_t *handle, alpm_list_t *packages)
 {
 	alpm_list_t *baddeps = NULL;
@@ -180,7 +210,9 @@ alpm_list_t *_alpm_innerconflicts(alpm_handle_t *handle, alpm_list_t *packages)
 	return baddeps;
 }
 
-/* Check for target vs (db - target) conflicts */
+/**
+ * @brief Returns a list of conflicts between a db and a list of packages.
+ */
 alpm_list_t *_alpm_outerconflicts(alpm_db_t *db, alpm_list_t *packages)
 {
 	alpm_list_t *baddeps = NULL;
@@ -202,10 +234,12 @@ alpm_list_t *_alpm_outerconflicts(alpm_db_t *db, alpm_list_t *packages)
 	return baddeps;
 }
 
-/** Check the package conflicts in a database
+/**
+ * @brief Check the package conflicts in a database
  *
  * @param handle the context handle
  * @param pkglist the list of packages to check
+ *
  * @return an alpm_list_t of alpm_conflict_t
  */
 alpm_list_t SYMEXPORT *alpm_checkconflicts(alpm_handle_t *handle,
@@ -215,69 +249,16 @@ alpm_list_t SYMEXPORT *alpm_checkconflicts(alpm_handle_t *handle,
 	return _alpm_innerconflicts(handle, pkglist);
 }
 
-static const int DIFFERENCE = 0;
-static const int INTERSECT = 1;
-/* Returns a set operation on the provided two lists of files.
- * Pre-condition: both lists are sorted!
- * When done, free the list but NOT the contained data.
+/**
+ * @brief Creates and adds a file conflict to a conflict list.
  *
- * Operations:
- *   DIFFERENCE - a difference operation is performed. filesA - filesB.
- *   INTERSECT - an intersection operation is performed. filesA & filesB.
- */
-static alpm_list_t *filelist_operation(alpm_filelist_t *filesA,
-		alpm_filelist_t *filesB, int operation)
-{
-	alpm_list_t *ret = NULL;
-	size_t ctrA = 0, ctrB = 0;
-
-	while(ctrA < filesA->count && ctrB < filesB->count) {
-		alpm_file_t *fileA = filesA->files + ctrA;
-		alpm_file_t *fileB = filesB->files + ctrB;
-		const char *strA = fileA->name;
-		const char *strB = fileB->name;
-		/* skip directories, we don't care about them */
-		if(strA[strlen(strA)-1] == '/') {
-			ctrA++;
-		} else if(strB[strlen(strB)-1] == '/') {
-			ctrB++;
-		} else {
-			int cmp = strcmp(strA, strB);
-			if(cmp < 0) {
-				if(operation == DIFFERENCE) {
-					/* item only in filesA, qualifies as a difference */
-					ret = alpm_list_add(ret, fileA);
-				}
-				ctrA++;
-			} else if(cmp > 0) {
-				ctrB++;
-			} else {
-				if(operation == INTERSECT) {
-					/* item in both, qualifies as an intersect */
-					ret = alpm_list_add(ret, fileA);
-				}
-				ctrA++;
-				ctrB++;
-			}
-	  }
-	}
-
-	/* if doing a difference, ensure we have completely emptied pA */
-	while(operation == DIFFERENCE && ctrA < filesA->count) {
-		alpm_file_t *fileA = filesA->files + ctrA;
-		const char *strA = fileA->name;
-		/* skip directories */
-		if(strA[strlen(strA)-1] != '/') {
-			ret = alpm_list_add(ret, fileA);
-		}
-		ctrA++;
-	}
-
-	return ret;
-}
-
-/* Adds alpm_fileconflict_t to a conflicts list. Pass the conflicts list, the
- * conflicting file path, and either two packages or one package and NULL.
+ * @param handle the context handle
+ * @param conflicts the list of conflicts to append to
+ * @param filestr the conflicting file path
+ * @param pkg1 package that wishes to install the file
+ * @param pkg2 package that currently owns the file, or NULL if unowned
+ *
+ * @return the updated conflict list
  */
 static alpm_list_t *add_fileconflict(alpm_handle_t *handle,
 		alpm_list_t *conflicts, const char *filestr,
@@ -306,6 +287,9 @@ error:
 	RET_ERR(handle, ALPM_ERR_MEMORY, conflicts);
 }
 
+/**
+ * @brief Frees a conflict and its members.
+ */
 void _alpm_fileconflict_free(alpm_fileconflict_t *conflict)
 {
 	FREE(conflict->ctarget);
@@ -314,29 +298,58 @@ void _alpm_fileconflict_free(alpm_fileconflict_t *conflict)
 	FREE(conflict);
 }
 
-const alpm_file_t *_alpm_filelist_contains(alpm_filelist_t *filelist,
-		const char *name)
-{
-	size_t i;
-	const alpm_file_t *file = filelist->files;
-	for(i = 0; i < filelist->count; i++) {
-		if(strcmp(file->name, name) == 0) {
-			return file;
-		}
-		file++;
-	}
-	return NULL;
-}
-
-static int dir_belongsto_pkg(const char *root, const char *dirpath,
+/**
+ * @brief Recursively checks if a package owns all subdirectories and files in
+ * a directory.
+ *
+ * @param handle the context handle
+ * @param dirpath path of the directory to check
+ * @param pkg package being checked against
+ *
+ * @return 1 if a package owns all subdirectories and files or a directory
+ * cannot be opened, 0 otherwise
+ */
+static int dir_belongsto_pkg(alpm_handle_t *handle, const char *dirpath,
 		alpm_pkg_t *pkg)
 {
+	alpm_list_t *i;
 	struct stat sbuf;
 	char path[PATH_MAX];
 	char abspath[PATH_MAX];
-	struct dirent *ent = NULL;
 	DIR *dir;
+	struct dirent *ent = NULL;
+	const char *root = handle->root;
 
+	/* check directory is actually in package - used for subdirectory checks */
+	_alpm_filelist_resolve(handle, alpm_pkg_get_files(pkg));
+	if(!alpm_filelist_contains(alpm_pkg_get_files(pkg), dirpath)) {
+		_alpm_log(handle, ALPM_LOG_DEBUG,
+				"directory %s not in package %s\n", dirpath, pkg->name);
+		return 0;
+	}
+
+	/* TODO: this is an overly strict check but currently pacman will not
+	 * overwrite a directory with a file (case 10/11 in add.c). Adjusting that
+	 * is not simple as even if the directory is being unowned by a conflicting
+	 * package, pacman does not sort this to ensure all required directory
+	 * "removals" happen before installation of file/symlink */
+
+	/* check that no other _installed_ package owns the directory */
+	for(i = _alpm_db_get_pkgcache(handle->db_local); i; i = i->next) {
+		if(pkg == i->data) {
+			continue;
+		}
+
+		_alpm_filelist_resolve(handle, alpm_pkg_get_files(i->data));
+		if(alpm_filelist_contains(alpm_pkg_get_files(i->data), dirpath)) {
+			_alpm_log(handle, ALPM_LOG_DEBUG,
+					"file %s also in package %s\n", dirpath,
+					((alpm_pkg_t*)i->data)->name);
+			return 0;
+		}
+	}
+
+	/* check all files in directory are owned by the package */
 	snprintf(abspath, PATH_MAX, "%s%s", root, dirpath);
 	dir = opendir(abspath);
 	if(dir == NULL) {
@@ -349,23 +362,26 @@ static int dir_belongsto_pkg(const char *root, const char *dirpath,
 		if(strcmp(name, ".") == 0 || strcmp(name, "..") == 0) {
 			continue;
 		}
-		snprintf(path, PATH_MAX, "%s/%s", dirpath, name);
+		snprintf(path, PATH_MAX, "%s%s", dirpath, name);
 		snprintf(abspath, PATH_MAX, "%s%s", root, path);
 		if(stat(abspath, &sbuf) != 0) {
 			continue;
 		}
 		if(S_ISDIR(sbuf.st_mode)) {
-			if(dir_belongsto_pkg(root, path, pkg)) {
+			if(dir_belongsto_pkg(handle, path, pkg)) {
 				continue;
 			} else {
 				closedir(dir);
 				return 0;
 			}
 		} else {
-			if(_alpm_filelist_contains(alpm_pkg_get_files(pkg), path)) {
+			_alpm_filelist_resolve(handle, alpm_pkg_get_files(pkg));
+			if(alpm_filelist_contains(alpm_pkg_get_files(pkg), path)) {
 				continue;
 			} else {
 				closedir(dir);
+				_alpm_log(handle, ALPM_LOG_DEBUG,
+						"unowned file %s found in directory\n", path);
 				return 0;
 			}
 		}
@@ -374,20 +390,38 @@ static int dir_belongsto_pkg(const char *root, const char *dirpath,
 	return 1;
 }
 
-/* Find file conflicts that may occur during the transaction with two checks:
- * 1: check every target against every target
- * 2: check every target against the filesystem */
+/**
+ * @brief Find file conflicts that may occur during the transaction.
+ *
+ * @details Performs two checks:
+ *   1. check every target against every target
+ *   2. check every target against the filesystem
+ *
+ * @param handle the context handle
+ * @param upgrade list of packages being installed
+ * @param rem list of packages being removed
+ *
+ * @return list of file conflicts
+ */
 alpm_list_t *_alpm_db_find_fileconflicts(alpm_handle_t *handle,
-		alpm_list_t *upgrade, alpm_list_t *remove)
+		alpm_list_t *upgrade, alpm_list_t *rem)
 {
 	alpm_list_t *i, *conflicts = NULL;
 	size_t numtargs = alpm_list_count(upgrade);
 	size_t current;
+	size_t rootlen;
 
 	if(!upgrade) {
 		return NULL;
 	}
 
+	rootlen = strlen(handle->root);
+
+	/* make sure all files to be installed have been resolved */
+	for(i = upgrade; i; i = i->next) {
+		_alpm_filelist_resolve(handle, alpm_pkg_get_files(i->data));
+	}
+
 	/* TODO this whole function needs a huge change, which hopefully will
 	 * be possible with real transactions. Right now we only do half as much
 	 * here as we do when we actually extract files in add.c with our 12
@@ -395,32 +429,49 @@ alpm_list_t *_alpm_db_find_fileconflicts(alpm_handle_t *handle,
 	for(current = 0, i = upgrade; i; i = i->next, current++) {
 		alpm_pkg_t *p1 = i->data;
 		alpm_list_t *j;
-		alpm_filelist_t tmpfiles;
+		alpm_list_t *tmpfiles = NULL;
 		alpm_pkg_t *dbpkg;
-		size_t filenum;
 
 		int percent = (current * 100) / numtargs;
 		PROGRESS(handle, ALPM_PROGRESS_CONFLICTS_START, "", percent,
 		         numtargs, current);
+
 		/* CHECK 1: check every target against every target */
 		_alpm_log(handle, ALPM_LOG_DEBUG, "searching for file conflicts: %s\n",
 				p1->name);
 		for(j = i->next; j; j = j->next) {
 			alpm_list_t *common_files;
 			alpm_pkg_t *p2 = j->data;
-			common_files = filelist_operation(alpm_pkg_get_files(p1),
-					alpm_pkg_get_files(p2), INTERSECT);
+
+			alpm_filelist_t *p1_files = alpm_pkg_get_files(p1);
+			alpm_filelist_t *p2_files = alpm_pkg_get_files(p2);
+
+			common_files = _alpm_filelist_intersection(p1_files, p2_files);
 
 			if(common_files) {
 				alpm_list_t *k;
 				char path[PATH_MAX];
 				for(k = common_files; k; k = k->next) {
-					alpm_file_t *file = k->data;
-					snprintf(path, PATH_MAX, "%s%s", handle->root, file->name);
+					char *filename = k->data;
+					snprintf(path, PATH_MAX, "%s%s", handle->root, filename);
+
+					/* can skip file-file conflicts when forced *
+					 * checking presence in p2_files detects dir-file or file-dir
+					 * conflicts as the path from p1 is returned */
+					if((handle->trans->flags & ALPM_TRANS_FLAG_FORCE) &&
+							alpm_filelist_contains(p2_files, filename)) {
+						_alpm_log(handle, ALPM_LOG_DEBUG,
+							"%s exists in both '%s' and '%s'\n", filename,
+							p1->name, p2->name);
+						_alpm_log(handle, ALPM_LOG_DEBUG,
+							"file-file conflict being forced\n");
+						continue;
+					}
+
 					conflicts = add_fileconflict(handle, conflicts, path, p1, p2);
 					if(handle->pm_errno == ALPM_ERR_MEMORY) {
 						FREELIST(conflicts);
-						FREELIST(common_files);
+						alpm_list_free(common_files);
 						return NULL;
 					}
 				}
@@ -439,30 +490,30 @@ alpm_list_t *_alpm_db_find_fileconflicts(alpm_handle_t *handle,
 		 * that the former list needs to be freed while the latter list should NOT
 		 * be freed. */
 		if(dbpkg) {
-			alpm_list_t *difference;
 			/* older ver of package currently installed */
-			difference = filelist_operation(alpm_pkg_get_files(p1),
-					alpm_pkg_get_files(dbpkg), DIFFERENCE);
-			tmpfiles.count = alpm_list_count(difference);
-			tmpfiles.files = alpm_list_to_array(difference, tmpfiles.count,
-					sizeof(alpm_file_t));
-			alpm_list_free(difference);
+			_alpm_filelist_resolve(handle, alpm_pkg_get_files(dbpkg));
+			tmpfiles = _alpm_filelist_difference(alpm_pkg_get_files(p1),
+					alpm_pkg_get_files(dbpkg));
 		} else {
 			/* no version of package currently installed */
-			tmpfiles = *alpm_pkg_get_files(p1);
+			alpm_filelist_t *fl = alpm_pkg_get_files(p1);
+			size_t filenum;
+			for(filenum = 0; filenum < fl->count; filenum++) {
+				tmpfiles = alpm_list_add(tmpfiles, fl->resolved_path[filenum]);
+			}
 		}
 
-		for(filenum = 0; filenum < tmpfiles.count; filenum++) {
-			alpm_file_t *file = tmpfiles.files + filenum;
-			const char *filestr = file->name;
+		for(j = tmpfiles; j; j = j->next) {
+			const char *filestr = j->data;
 			const char *relative_path;
 			alpm_list_t *k;
 			/* have we acted on this conflict? */
 			int resolved_conflict = 0;
 			struct stat lsbuf;
 			char path[PATH_MAX];
+			size_t pathlen;
 
-			snprintf(path, PATH_MAX, "%s%s", handle->root, filestr);
+			pathlen = snprintf(path, PATH_MAX, "%s%s", handle->root, filestr);
 
 			/* stat the file - if it exists, do some checks */
 			if(_alpm_lstat(path, &lsbuf) != 0) {
@@ -471,7 +522,7 @@ alpm_list_t *_alpm_db_find_fileconflicts(alpm_handle_t *handle,
 
 			_alpm_log(handle, ALPM_LOG_DEBUG, "checking possible conflict: %s\n", path);
 
-			if(S_ISDIR(file->mode)) {
+			if(filestr[strlen(filestr) - 1] == '/') {
 				struct stat sbuf;
 				if(S_ISDIR(lsbuf.st_mode)) {
 					_alpm_log(handle, ALPM_LOG_DEBUG, "file is a directory, not a conflict\n");
@@ -486,15 +537,16 @@ alpm_list_t *_alpm_db_find_fileconflicts(alpm_handle_t *handle,
 				/* if we made it to here, we want all subsequent path comparisons to
 				 * not include the trailing slash. This allows things like file ->
 				 * directory replacements. */
-				path[strlen(path) - 1] = '\0';
+				path[pathlen - 1] = '\0';
 			}
 
-			relative_path = path + strlen(handle->root);
+			relative_path = path + rootlen;
 
 			/* Check remove list (will we remove the conflicting local file?) */
-			for(k = remove; k && !resolved_conflict; k = k->next) {
+			for(k = rem; k && !resolved_conflict; k = k->next) {
 				alpm_pkg_t *rempkg = k->data;
-				if(rempkg && _alpm_filelist_contains(alpm_pkg_get_files(rempkg),
+				_alpm_filelist_resolve(handle, alpm_pkg_get_files(rempkg));
+				if(rempkg && alpm_filelist_contains(alpm_pkg_get_files(rempkg),
 							relative_path)) {
 					_alpm_log(handle, ALPM_LOG_DEBUG,
 							"local file will be removed, not a conflict\n");
@@ -511,7 +563,8 @@ alpm_list_t *_alpm_db_find_fileconflicts(alpm_handle_t *handle,
 				alpm_pkg_t *localp2 = _alpm_db_get_pkgfromcache(handle->db_local, p2->name);
 
 				/* localp2->files will be removed (target conflicts are handled by CHECK 1) */
-				if(localp2 && _alpm_filelist_contains(alpm_pkg_get_files(localp2), filestr)) {
+				_alpm_filelist_resolve(handle, alpm_pkg_get_files(localp2));
+				if(localp2 && alpm_filelist_contains(alpm_pkg_get_files(localp2), filestr)) {
 					/* skip removal of file, but not add. this will prevent a second
 					 * package from removing the file when it was already installed
 					 * by its new owner (whether the file is in backup array or not */
@@ -527,11 +580,11 @@ alpm_list_t *_alpm_db_find_fileconflicts(alpm_handle_t *handle,
 			if(!resolved_conflict && S_ISDIR(lsbuf.st_mode) && dbpkg) {
 				char *dir = malloc(strlen(filestr) + 2);
 				sprintf(dir, "%s/", filestr);
-				if(_alpm_filelist_contains(alpm_pkg_get_files(dbpkg), dir)) {
+				if(alpm_filelist_contains(alpm_pkg_get_files(dbpkg), dir)) {
 					_alpm_log(handle, ALPM_LOG_DEBUG,
-							"check if all files in %s belongs to %s\n",
+							"checking if all files in %s belong to %s\n",
 							dir, dbpkg->name);
-					resolved_conflict = dir_belongsto_pkg(handle->root, filestr, dbpkg);
+					resolved_conflict = dir_belongsto_pkg(handle, dir, dbpkg);
 				}
 				free(dir);
 			}
@@ -541,17 +594,15 @@ alpm_list_t *_alpm_db_find_fileconflicts(alpm_handle_t *handle,
 			 * consideration cannot itself be a link, as it might be unowned- path
 			 * components can be safely checked as all directories are "unowned". */
 			if(!resolved_conflict && dbpkg && !S_ISLNK(lsbuf.st_mode)) {
-				char *rpath = calloc(PATH_MAX, sizeof(char));
-				const char *relative_rpath;
+				char rpath[PATH_MAX];
 				if(realpath(path, rpath)) {
-					relative_rpath = rpath + strlen(handle->root);
-					if(_alpm_filelist_contains(alpm_pkg_get_files(dbpkg), relative_rpath)) {
+					const char *relative_rpath = rpath + rootlen;
+					if(alpm_filelist_contains(alpm_pkg_get_files(dbpkg), relative_rpath)) {
 						_alpm_log(handle, ALPM_LOG_DEBUG,
 								"package contained the resolved realpath\n");
 						resolved_conflict = 1;
 					}
 				}
-				free(rpath);
 			}
 
 			/* is the file unowned and in the backup list of the new package? */
@@ -559,7 +610,8 @@ alpm_list_t *_alpm_db_find_fileconflicts(alpm_handle_t *handle,
 				alpm_list_t *local_pkgs = _alpm_db_get_pkgcache(handle->db_local);
 				int found = 0;
 				for(k = local_pkgs; k && !found; k = k->next) {
-					if(_alpm_filelist_contains(alpm_pkg_get_files(k->data), filestr)) {
+					_alpm_filelist_resolve(handle, alpm_pkg_get_files(k->data));
+					if(alpm_filelist_contains(alpm_pkg_get_files(k->data), filestr)) {
 							found = 1;
 					}
 				}
@@ -570,22 +622,24 @@ alpm_list_t *_alpm_db_find_fileconflicts(alpm_handle_t *handle,
 				}
 			}
 
+			/* skip file-file conflicts when being forced */
+			if((handle->trans->flags & ALPM_TRANS_FLAG_FORCE) &&
+					!S_ISDIR(lsbuf.st_mode)) {
+				_alpm_log(handle, ALPM_LOG_DEBUG,
+							"conflict with file on filesystem being forced\n");
+				resolved_conflict = 1;
+			}
+
 			if(!resolved_conflict) {
 				conflicts = add_fileconflict(handle, conflicts, path, p1, NULL);
 				if(handle->pm_errno == ALPM_ERR_MEMORY) {
 					FREELIST(conflicts);
-					if(dbpkg) {
-						/* only freed if it was generated from filelist_operation() */
-						free(tmpfiles.files);
-					}
+					alpm_list_free(tmpfiles);
 					return NULL;
 				}
 			}
 		}
-		if(dbpkg) {
-			/* only freed if it was generated from filelist_operation() */
-			free(tmpfiles.files);
-		}
+		alpm_list_free(tmpfiles);
 	}
 	PROGRESS(handle, ALPM_PROGRESS_CONFLICTS_START, "", 100,
 			numtargs, current);
