@@ -1,7 +1,7 @@
 /*
  *  deps.c
  *
- *  Copyright (c) 2006-2011 Pacman Development Team <pacman-dev@archlinux.org>
+ *  Copyright (c) 2006-2013 Pacman Development Team <pacman-dev@archlinux.org>
  *  Copyright (c) 2002-2006 by Judd Vinet <jvinet@zeroflux.org>
  *  Copyright (c) 2005 by Aurelien Foret <orelien@chez.com>
  *  Copyright (c) 2005, 2006 by Miklos Vajna <vmiklos@frugalware.org>
@@ -20,8 +20,6 @@
  *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
  */
 
-#include "config.h"
-
 #include <stdlib.h>
 #include <stdio.h>
 #include <string.h>
@@ -41,6 +39,7 @@ void _alpm_dep_free(alpm_depend_t *dep)
 {
 	FREE(dep->name);
 	FREE(dep->version);
+	FREE(dep->desc);
 	FREE(dep);
 }
 
@@ -159,15 +158,17 @@ alpm_list_t *_alpm_sortbydeps(alpm_handle_t *handle,
 			else if(nextchild->state == -1) {
 				alpm_pkg_t *vertexpkg = vertex->data;
 				alpm_pkg_t *childpkg = nextchild->data;
-				const char *message;
 
 				_alpm_log(handle, ALPM_LOG_WARNING, _("dependency cycle detected:\n"));
 				if(reverse) {
-					message =_("%s will be removed after its %s dependency\n");
+					_alpm_log(handle, ALPM_LOG_WARNING,
+							_("%s will be removed after its %s dependency\n"),
+							vertexpkg->name, childpkg->name);
 				} else {
-					message =_("%s will be installed before its %s dependency\n");
+					_alpm_log(handle, ALPM_LOG_WARNING,
+							_("%s will be installed before its %s dependency\n"),
+							vertexpkg->name, childpkg->name);
 				}
-				_alpm_log(handle, ALPM_LOG_WARNING, message, vertexpkg->name, childpkg->name);
 			}
 		}
 		if(!found) {
@@ -204,8 +205,10 @@ alpm_list_t *_alpm_sortbydeps(alpm_handle_t *handle,
 
 static int no_dep_version(alpm_handle_t *handle)
 {
-	int flags = alpm_trans_get_flags(handle);
-	return flags != -1 && (flags & ALPM_TRANS_FLAG_NODEPVERSION);
+	if(!handle->trans) {
+		return 0;
+	}
+	return (handle->trans->flags & ALPM_TRANS_FLAG_NODEPVERSION);
 }
 
 static alpm_depend_t *filtered_depend(alpm_depend_t *dep, int nodepversion)
@@ -266,7 +269,7 @@ alpm_pkg_t SYMEXPORT *alpm_find_satisfier(alpm_list_t *pkgs, const char *depstri
  * @return an alpm_list_t* of alpm_depmissing_t pointers.
  */
 alpm_list_t SYMEXPORT *alpm_checkdeps(alpm_handle_t *handle,
-		alpm_list_t *pkglist, alpm_list_t *remove, alpm_list_t *upgrade,
+		alpm_list_t *pkglist, alpm_list_t *rem, alpm_list_t *upgrade,
 		int reversedeps)
 {
 	alpm_list_t *i, *j;
@@ -278,7 +281,7 @@ alpm_list_t SYMEXPORT *alpm_checkdeps(alpm_handle_t *handle,
 
 	for(i = pkglist; i; i = i->next) {
 		alpm_pkg_t *pkg = i->data;
-		if(_alpm_pkg_find(remove, pkg->name) || _alpm_pkg_find(upgrade, pkg->name)) {
+		if(alpm_pkg_find(rem, pkg->name) || alpm_pkg_find(upgrade, pkg->name)) {
 			modified = alpm_list_add(modified, pkg);
 		} else {
 			dblist = alpm_list_add(dblist, pkg);
@@ -395,7 +398,7 @@ int _alpm_depcmp(alpm_pkg_t *pkg, alpm_depend_t *dep)
 			/* any version will satisfy the requirement */
 			satisfy = (provision->name_hash == dep->name_hash
 					&& strcmp(provision->name, dep->name) == 0);
-		} else if (provision->mod == ALPM_DEP_MOD_EQ) {
+		} else if(provision->mod == ALPM_DEP_MOD_EQ) {
 			/* provision specifies a version, so try it out */
 			satisfy = (provision->name_hash == dep->name_hash
 					&& strcmp(provision->name, dep->name) == 0
@@ -409,7 +412,7 @@ int _alpm_depcmp(alpm_pkg_t *pkg, alpm_depend_t *dep)
 alpm_depend_t *_alpm_splitdep(const char *depstring)
 {
 	alpm_depend_t *depend;
-	const char *ptr, *version;
+	const char *ptr, *version, *desc;
 	size_t deplen;
 
 	if(depstring == NULL) {
@@ -417,7 +420,17 @@ alpm_depend_t *_alpm_splitdep(const char *depstring)
 	}
 
 	MALLOC(depend, sizeof(alpm_depend_t), return NULL);
-	deplen = strlen(depstring);
+
+	/* Note the extra space in ": " to avoid matching the epoch */
+	if((desc = strstr(depstring, ": ")) != NULL) {
+		STRDUP(depend->desc, desc + 2, return NULL);
+		deplen = desc - depstring;
+	} else {
+		/* no description- point desc at NULL at end of string for later use */
+		depend->desc = NULL;
+		deplen = strlen(depstring);
+		desc = depstring + deplen;
+	}
 
 	/* Find a version comparator if one exists. If it does, set the type and
 	 * increment the ptr accordingly so we can copy the right strings. */
@@ -442,21 +455,18 @@ alpm_depend_t *_alpm_splitdep(const char *depstring)
 		depend->mod = ALPM_DEP_MOD_EQ;
 		version = ptr + 1;
 	} else {
-		/* no version specified, leave ptr NULL and set version to NULL */
+		/* no version specified, set ptr to end of string and version to NULL */
+		ptr = depstring + deplen;
 		depend->mod = ALPM_DEP_MOD_ANY;
 		depend->version = NULL;
 		version = NULL;
 	}
 
 	/* copy the right parts to the right places */
-	if(ptr) {
-		STRNDUP(depend->name, depstring, ptr - depstring, return NULL);
-	} else {
-		STRDUP(depend->name, depstring, return NULL);
-	}
+	STRNDUP(depend->name, depstring, ptr - depstring, return NULL);
 	depend->name_hash = _alpm_hash_sdbm(depend->name);
 	if(version) {
-		STRDUP(depend->version, version, return NULL);
+		STRNDUP(depend->version, version, desc - version, return NULL);
 	}
 
 	return depend;
@@ -468,8 +478,9 @@ alpm_depend_t *_alpm_dep_dup(const alpm_depend_t *dep)
 	CALLOC(newdep, 1, sizeof(alpm_depend_t), return NULL);
 
 	STRDUP(newdep->name, dep->name, return NULL);
-	newdep->name_hash = dep->name_hash;
 	STRDUP(newdep->version, dep->version, return NULL);
+	STRDUP(newdep->desc, dep->desc, return NULL);
+	newdep->name_hash = dep->name_hash;
 	newdep->mod = dep->mod;
 
 	return newdep;
@@ -484,7 +495,7 @@ static int can_remove_package(alpm_db_t *db, alpm_pkg_t *pkg,
 {
 	alpm_list_t *i;
 
-	if(_alpm_pkg_find(targets, pkg->name)) {
+	if(alpm_pkg_find(targets, pkg->name)) {
 		return 0;
 	}
 
@@ -506,7 +517,7 @@ static int can_remove_package(alpm_db_t *db, alpm_pkg_t *pkg,
 	/* see if other packages need it */
 	for(i = _alpm_db_get_pkgcache(db); i; i = i->next) {
 		alpm_pkg_t *lpkg = i->data;
-		if(_alpm_dep_edge(lpkg, pkg) && !_alpm_pkg_find(targets, lpkg->name)) {
+		if(_alpm_dep_edge(lpkg, pkg) && !alpm_pkg_find(targets, lpkg->name)) {
 			return 0;
 		}
 	}
@@ -580,7 +591,7 @@ static alpm_pkg_t *resolvedep(alpm_handle_t *handle, alpm_depend_t *dep,
 	for(i = dbs; i; i = i->next) {
 		alpm_pkg_t *pkg = _alpm_db_get_pkgfromcache(i->data, dep->name);
 		if(pkg && _alpm_depcmp_literal(pkg, dep)
-				&& !_alpm_pkg_find(excluding, pkg->name)) {
+				&& !alpm_pkg_find(excluding, pkg->name)) {
 			if(_alpm_pkg_should_ignore(handle, pkg)) {
 				int install = 0;
 				if(prompt) {
@@ -605,7 +616,7 @@ static alpm_pkg_t *resolvedep(alpm_handle_t *handle, alpm_depend_t *dep,
 			/* with hash != hash, we can even skip the strcmp() as we know they can't
 			 * possibly be the same string */
 			if(pkg->name_hash != dep->name_hash && _alpm_depcmp(pkg, dep)
-					&& !_alpm_pkg_find(excluding, pkg->name)) {
+					&& !alpm_pkg_find(excluding, pkg->name)) {
 				if(_alpm_pkg_should_ignore(handle, pkg)) {
 					int install = 0;
 					if(prompt) {
@@ -639,14 +650,14 @@ static alpm_pkg_t *resolvedep(alpm_handle_t *handle, alpm_depend_t *dep,
 	count = alpm_list_count(providers);
 	if(count >= 1) {
 		/* default to first provider if there is no QUESTION callback */
-		int index = 0;
+		int idx = 0;
 		if(count > 1) {
 			/* if there is more than one provider, we ask the user */
 			QUESTION(handle, ALPM_QUESTION_SELECT_PROVIDER,
-					providers, dep, NULL, &index);
+					providers, dep, NULL, &idx);
 		}
-		if(index >= 0 && index < count) {
-			alpm_list_t *nth = alpm_list_nth(providers, index);
+		if(idx >= 0 && idx < count) {
+			alpm_list_t *nth = alpm_list_nth(providers, idx);
 			alpm_pkg_t *pkg = nth->data;
 			alpm_list_free(providers);
 			return pkg;
@@ -711,15 +722,15 @@ alpm_pkg_t SYMEXPORT *alpm_find_dbs_satisfier(alpm_handle_t *handle,
  */
 int _alpm_resolvedeps(alpm_handle_t *handle, alpm_list_t *localpkgs,
 		alpm_pkg_t *pkg, alpm_list_t *preferred, alpm_list_t **packages,
-		alpm_list_t *remove, alpm_list_t **data)
+		alpm_list_t *rem, alpm_list_t **data)
 {
 	int ret = 0;
-	alpm_list_t *i, *j;
+	alpm_list_t *j;
 	alpm_list_t *targ;
 	alpm_list_t *deps = NULL;
 	alpm_list_t *packages_copy;
 
-	if(_alpm_pkg_find(*packages, pkg->name) != NULL) {
+	if(alpm_pkg_find(*packages, pkg->name) != NULL) {
 		return 0;
 	}
 
@@ -731,49 +742,49 @@ int _alpm_resolvedeps(alpm_handle_t *handle, alpm_list_t *localpkgs,
 	*packages = alpm_list_add(*packages, pkg);
 
 	_alpm_log(handle, ALPM_LOG_DEBUG, "started resolving dependencies\n");
-	for(i = alpm_list_last(*packages); i; i = i->next) {
-		alpm_pkg_t *tpkg = i->data;
-		targ = alpm_list_add(NULL, tpkg);
-		deps = alpm_checkdeps(handle, localpkgs, remove, targ, 0);
-		alpm_list_free(targ);
+	targ = alpm_list_add(NULL, pkg);
+	deps = alpm_checkdeps(handle, localpkgs, rem, targ, 0);
+	alpm_list_free(targ);
+	targ = NULL;
 
-		for(j = deps; j; j = j->next) {
-			alpm_depmissing_t *miss = j->data;
-			alpm_depend_t *missdep = miss->depend;
-			/* check if one of the packages in the [*packages] list already satisfies
-			 * this dependency */
-			if(find_dep_satisfier(*packages, missdep)) {
-				_alpm_depmiss_free(miss);
-				continue;
+	for(j = deps; j; j = j->next) {
+		alpm_depmissing_t *miss = j->data;
+		alpm_depend_t *missdep = miss->depend;
+		/* check if one of the packages in the [*packages] list already satisfies
+		 * this dependency */
+		if(find_dep_satisfier(*packages, missdep)) {
+			_alpm_depmiss_free(miss);
+			continue;
+		}
+		/* check if one of the packages in the [preferred] list already satisfies
+		 * this dependency */
+		alpm_pkg_t *spkg = find_dep_satisfier(preferred, missdep);
+		if(!spkg) {
+			/* find a satisfier package in the given repositories */
+			spkg = resolvedep(handle, missdep, handle->dbs_sync, *packages, 0);
+		}
+		if(spkg && _alpm_resolvedeps(handle, localpkgs, spkg, preferred, packages, rem, data) == 0) {
+			_alpm_log(handle, ALPM_LOG_DEBUG,
+					"pulling dependency %s (needed by %s)\n",
+					spkg->name, pkg->name);
+			_alpm_depmiss_free(miss);
+		} else if(resolvedep(handle, missdep, (targ = alpm_list_add(NULL, handle->db_local)), rem, 0)) {
+			_alpm_depmiss_free(miss);
+		} else {
+			handle->pm_errno = ALPM_ERR_UNSATISFIED_DEPS;
+			char *missdepstring = alpm_dep_compute_string(missdep);
+			_alpm_log(handle, ALPM_LOG_WARNING,
+					_("cannot resolve \"%s\", a dependency of \"%s\"\n"),
+					missdepstring, pkg->name);
+			free(missdepstring);
+			if(data) {
+				*data = alpm_list_add(*data, miss);
 			}
-			/* check if one of the packages in the [preferred] list already satisfies
-			 * this dependency */
-			alpm_pkg_t *spkg = find_dep_satisfier(preferred, missdep);
-			if(!spkg) {
-				/* find a satisfier package in the given repositories */
-				spkg = resolvedep(handle, missdep, handle->dbs_sync, *packages, 0);
-			}
-			if(!spkg) {
-				handle->pm_errno = ALPM_ERR_UNSATISFIED_DEPS;
-				char *missdepstring = alpm_dep_compute_string(missdep);
-				_alpm_log(handle, ALPM_LOG_WARNING,
-						_("cannot resolve \"%s\", a dependency of \"%s\"\n"),
-						missdepstring, tpkg->name);
-				free(missdepstring);
-				if(data) {
-					*data = alpm_list_add(*data, miss);
-				}
-				ret = -1;
-			} else {
-				_alpm_log(handle, ALPM_LOG_DEBUG,
-						"pulling dependency %s (needed by %s)\n",
-						spkg->name, tpkg->name);
-				*packages = alpm_list_add(*packages, spkg);
-				_alpm_depmiss_free(miss);
-			}
+			ret = -1;
 		}
-		alpm_list_free(deps);
 	}
+	alpm_list_free(targ);
+	alpm_list_free(deps);
 
 	if(ret != 0) {
 		alpm_list_free(*packages);
@@ -792,7 +803,7 @@ int _alpm_resolvedeps(alpm_handle_t *handle, alpm_list_t *localpkgs,
  */
 char SYMEXPORT *alpm_dep_compute_string(const alpm_depend_t *dep)
 {
-	const char *name, *opr, *ver;
+	const char *name, *opr, *ver, *desc_delim, *desc;
 	char *str;
 	size_t len;
 
@@ -834,12 +845,21 @@ char SYMEXPORT *alpm_dep_compute_string(const alpm_depend_t *dep)
 		ver = "";
 	}
 
+	if(dep->desc) {
+		desc_delim = ": ";
+		desc = dep->desc;
+	} else {
+		desc_delim = "";
+		desc = "";
+	}
+
 	/* we can always compute len and print the string like this because opr
 	 * and ver will be empty when ALPM_DEP_MOD_ANY is the depend type. the
 	 * reassignments above also ensure we do not do a strlen(NULL). */
-	len = strlen(name) + strlen(opr) + strlen(ver) + 1;
+	len = strlen(name) + strlen(opr) + strlen(ver)
+		+ strlen(desc_delim) + strlen(desc) + 1;
 	MALLOC(str, len, return NULL);
-	snprintf(str, len, "%s%s%s", name, opr, ver);
+	snprintf(str, len, "%s%s%s%s%s", name, opr, ver, desc_delim, desc);
 
 	return str;
 }
