@@ -1,7 +1,7 @@
 /*
  *  download.c
  *
- *  Copyright (c) 2006-2012 Pacman Development Team <pacman-dev@archlinux.org>
+ *  Copyright (c) 2006-2013 Pacman Development Team <pacman-dev@archlinux.org>
  *  Copyright (c) 2002-2006 by Judd Vinet <jvinet@zeroflux.org>
  *
  *  This program is free software; you can redistribute it and/or modify
@@ -18,8 +18,6 @@
  *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
  */
 
-#include "config.h"
-
 #include <stdlib.h>
 #include <stdio.h>
 #include <errno.h>
@@ -98,6 +96,11 @@ static int dload_progress_cb(void *file, double dltotal, double dlnow,
 	struct dload_payload *payload = (struct dload_payload *)file;
 	off_t current_size, total_size;
 
+	/* avoid displaying progress bar for redirects with a body */
+	if(payload->respcode >= 300) {
+		return 0;
+	}
+
 	/* SIGINT sent, abort by alerting curl */
 	if(dload_interrupted) {
 		return 1;
@@ -201,6 +204,7 @@ static size_t dload_parseheader_cb(void *ptr, size_t size, size_t nmemb, void *u
 	const char * const cd_header = "Content-Disposition:";
 	const char * const fn_key = "filename=";
 	struct dload_payload *payload = (struct dload_payload *)user;
+	long respcode;
 
 	if(_alpm_raw_ncmp(cd_header, ptr, strlen(cd_header)) == 0) {
 		if((fptr = strstr(ptr, fn_key))) {
@@ -222,6 +226,11 @@ static size_t dload_parseheader_cb(void *ptr, size_t size, size_t nmemb, void *u
 		}
 	}
 
+	curl_easy_getinfo(payload->handle->curl, CURLINFO_RESPONSE_CODE, &respcode);
+	if(payload->respcode != respcode) {
+		payload->respcode = respcode;
+	}
+
 	return realsize;
 }
 
@@ -285,7 +294,7 @@ static void curl_set_handle_opts(struct dload_payload *payload,
 	curl_easy_setopt(curl, CURLOPT_FOLLOWLOCATION, 1L);
 	curl_easy_setopt(curl, CURLOPT_PROGRESSFUNCTION, dload_progress_cb);
 	curl_easy_setopt(curl, CURLOPT_PROGRESSDATA, (void *)payload);
-	curl_easy_setopt(curl, CURLOPT_LOW_SPEED_LIMIT, 1024L);
+	curl_easy_setopt(curl, CURLOPT_LOW_SPEED_LIMIT, 1L);
 	curl_easy_setopt(curl, CURLOPT_LOW_SPEED_TIME, 10L);
 	curl_easy_setopt(curl, CURLOPT_HEADERFUNCTION, dload_parseheader_cb);
 	curl_easy_setopt(curl, CURLOPT_WRITEHEADER, (void *)payload);
@@ -324,7 +333,7 @@ static void curl_set_handle_opts(struct dload_payload *payload,
 	}
 }
 
-static void mask_signal(int signal, void (*handler)(int),
+static void mask_signal(int signum, void (*handler)(int),
 		struct sigaction *origaction)
 {
 	struct sigaction newaction;
@@ -333,13 +342,13 @@ static void mask_signal(int signal, void (*handler)(int),
 	sigemptyset(&newaction.sa_mask);
 	newaction.sa_flags = 0;
 
-	sigaction(signal, NULL, origaction);
-	sigaction(signal, &newaction, NULL);
+	sigaction(signum, NULL, origaction);
+	sigaction(signum, &newaction, NULL);
 }
 
-static void unmask_signal(int signal, struct sigaction sa)
+static void unmask_signal(int signum, struct sigaction *sa)
 {
-	sigaction(signal, &sa, NULL);
+	sigaction(signum, sa, NULL);
 }
 
 static FILE *create_tempfile(struct dload_payload *payload, const char *localpath)
@@ -357,11 +366,10 @@ static FILE *create_tempfile(struct dload_payload *payload, const char *localpat
 			fchmod(fd, ~(_getumask()) & 0666) ||
 			!(fp = fdopen(fd, payload->tempfile_openmode))) {
 		unlink(randpath);
-		if(fd >= 0) {
-			close(fd);
-		}
+		CLOSE(fd);
 		_alpm_log(payload->handle, ALPM_LOG_ERROR,
 				_("failed to create temporary file for download\n"));
+		free(randpath);
 		return NULL;
 	}
 	/* fp now points to our alpmtmp.XXXXXX */
@@ -378,7 +386,7 @@ static FILE *create_tempfile(struct dload_payload *payload, const char *localpat
 #define HOSTNAME_SIZE 256
 
 static int curl_download_internal(struct dload_payload *payload,
-		const char *localpath, char **final_file)
+		const char *localpath, char **final_file, char **final_url)
 {
 	int ret = -1;
 	FILE *localf = NULL;
@@ -386,7 +394,7 @@ static int curl_download_internal(struct dload_payload *payload,
 	char hostname[HOSTNAME_SIZE];
 	char error_buffer[CURL_ERROR_SIZE] = {0};
 	struct stat st;
-	long timecond, respcode = 0, remote_time = -1;
+	long timecond, remote_time = -1;
 	double remote_size, bytes_dl;
 	struct sigaction orig_sig_pipe, orig_sig_int;
 	/* shortcut to our handle within the payload */
@@ -409,7 +417,8 @@ static int curl_download_internal(struct dload_payload *payload,
 		RET_ERR(handle, ALPM_ERR_SERVER_BAD_URL, -1);
 	}
 
-	if(strlen(payload->remote_name) > 0 && strcmp(payload->remote_name, ".sig") != 0) {
+	if(payload->remote_name && strlen(payload->remote_name) > 0 &&
+			strcmp(payload->remote_name, ".sig") != 0) {
 		payload->destfile_name = get_fullpath(localpath, payload->remote_name, "");
 		payload->tempfile_name = get_fullpath(localpath, payload->remote_name, ".part");
 		if(!payload->destfile_name || !payload->tempfile_name) {
@@ -431,6 +440,10 @@ static int curl_download_internal(struct dload_payload *payload,
 	if(localf == NULL) {
 		localf = fopen(payload->tempfile_name, payload->tempfile_openmode);
 		if(localf == NULL) {
+			handle->pm_errno = ALPM_ERR_RETRIEVE;
+			_alpm_log(handle, ALPM_LOG_ERROR,
+					_("could not open file %s: %s\n"),
+					payload->tempfile_name, strerror(errno));
 			goto cleanup;
 		}
 	}
@@ -461,13 +474,12 @@ static int curl_download_internal(struct dload_payload *payload,
 	switch(payload->curlerr) {
 		case CURLE_OK:
 			/* get http/ftp response code */
-			curl_easy_getinfo(curl, CURLINFO_RESPONSE_CODE, &respcode);
-			_alpm_log(handle, ALPM_LOG_DEBUG, "response code: %ld\n", respcode);
-			if(respcode >= 400) {
+			_alpm_log(handle, ALPM_LOG_DEBUG, "response code: %ld\n", payload->respcode);
+			if(payload->respcode >= 400) {
 				payload->unlink_on_fail = 1;
 				/* non-translated message is same as libcurl */
 				snprintf(error_buffer, sizeof(error_buffer),
-						"The requested URL returned error: %ld", respcode);
+						"The requested URL returned error: %ld", payload->respcode);
 				_alpm_log(handle, ALPM_LOG_ERROR,
 						_("failed retrieving file '%s' from %s : %s\n"),
 						payload->remote_name, hostname, error_buffer);
@@ -511,6 +523,10 @@ static int curl_download_internal(struct dload_payload *payload,
 	curl_easy_getinfo(curl, CURLINFO_CONDITION_UNMET, &timecond);
 	curl_easy_getinfo(curl, CURLINFO_EFFECTIVE_URL, &effective_url);
 
+	if(final_url != NULL) {
+		*final_url = effective_url;
+	}
+
 	/* time condition was met and we didn't download anything. we need to
 	 * clean up the 0 byte .part file that's left behind. */
 	if(timecond == 1 && DOUBLE_EQ(bytes_dl, 0)) {
@@ -582,8 +598,8 @@ cleanup:
 	}
 
 	/* restore the old signal handlers */
-	unmask_signal(SIGINT, orig_sig_int);
-	unmask_signal(SIGPIPE, orig_sig_pipe);
+	unmask_signal(SIGINT, &orig_sig_int);
+	unmask_signal(SIGPIPE, &orig_sig_pipe);
 	/* if we were interrupted, trip the old handler */
 	if(dload_interrupted) {
 		raise(SIGINT);
@@ -601,13 +617,13 @@ cleanup:
  * @return 0 on success, -1 on error (pm_errno is set accordingly if errors_ok == 0)
  */
 int _alpm_download(struct dload_payload *payload, const char *localpath,
-		char **final_file)
+		char **final_file, char **final_url)
 {
 	alpm_handle_t *handle = payload->handle;
 
 	if(handle->fetchcb == NULL) {
 #ifdef HAVE_LIBCURL
-		return curl_download_internal(payload, localpath, final_file);
+		return curl_download_internal(payload, localpath, final_file, final_url);
 #else
 		RET_ERR(handle, ALPM_ERR_EXTERNAL_DOWNLOAD, -1);
 #endif
@@ -622,18 +638,18 @@ int _alpm_download(struct dload_payload *payload, const char *localpath,
 
 static char *filecache_find_url(alpm_handle_t *handle, const char *url)
 {
-	const char *basename = strrchr(url, '/');
+	const char *filebase = strrchr(url, '/');
 
-	if(basename == NULL) {
+	if(filebase == NULL) {
 		return NULL;
 	}
 
-	basename++;
-	if(basename == '\0') {
+	filebase++;
+	if(filebase == '\0') {
 		return NULL;
 	}
 
-	return _alpm_filecache_find(handle, basename);
+	return _alpm_filecache_find(handle, filebase);
 }
 
 /** Fetch a remote pkg. */
@@ -641,7 +657,7 @@ char SYMEXPORT *alpm_fetch_pkgurl(alpm_handle_t *handle, const char *url)
 {
 	char *filepath;
 	const char *cachedir;
-	char *final_file = NULL;
+	char *final_file = NULL, *final_pkg_url = NULL;
 	struct dload_payload payload;
 	int ret = 0;
 
@@ -661,7 +677,7 @@ char SYMEXPORT *alpm_fetch_pkgurl(alpm_handle_t *handle, const char *url)
 		payload.handle = handle;
 
 		/* download the file */
-		ret = _alpm_download(&payload, cachedir, &final_file);
+		ret = _alpm_download(&payload, cachedir, &final_file, &final_pkg_url);
 		_alpm_dload_payload_reset(&payload);
 		if(ret == -1) {
 			_alpm_log(handle, ALPM_LOG_WARNING, _("failed to download %s\n"), url);
@@ -672,13 +688,13 @@ char SYMEXPORT *alpm_fetch_pkgurl(alpm_handle_t *handle, const char *url)
 	}
 
 	/* attempt to download the signature */
-	if(ret == 0 && (handle->siglevel & ALPM_SIG_PACKAGE)) {
+	if(ret == 0 && final_pkg_url && (handle->siglevel & ALPM_SIG_PACKAGE)) {
 		char *sig_filepath, *sig_final_file = NULL;
 		size_t len;
 
-		len = strlen(url) + 5;
+		len = strlen(final_pkg_url) + 5;
 		MALLOC(payload.fileurl, len, RET_ERR(handle, ALPM_ERR_MEMORY, NULL));
-		snprintf(payload.fileurl, len, "%s.sig", url);
+		snprintf(payload.fileurl, len, "%s.sig", final_pkg_url);
 
 		sig_filepath = filecache_find_url(handle, payload.fileurl);
 		if(sig_filepath == NULL) {
@@ -689,7 +705,7 @@ char SYMEXPORT *alpm_fetch_pkgurl(alpm_handle_t *handle, const char *url)
 			/* set hard upper limit of 16KiB */
 			payload.max_size = 16 * 1024;
 
-			ret = _alpm_download(&payload, cachedir, &sig_final_file);
+			ret = _alpm_download(&payload, cachedir, &sig_final_file, NULL);
 			if(ret == -1 && !payload.errors_ok) {
 				_alpm_log(handle, ALPM_LOG_WARNING,
 						_("failed to download %s\n"), payload.fileurl);
@@ -723,6 +739,7 @@ void _alpm_dload_payload_reset(struct dload_payload *payload)
 	FREE(payload->destfile_name);
 	FREE(payload->content_disp_name);
 	FREE(payload->fileurl);
+	memset(payload, '\0', sizeof(*payload));
 }
 
 /* vim: set ts=2 sw=2 noet: */
