@@ -1,7 +1,7 @@
 /*
  *  package.c
  *
- *  Copyright (c) 2006-2011 Pacman Development Team <pacman-dev@archlinux.org>
+ *  Copyright (c) 2006-2013 Pacman Development Team <pacman-dev@archlinux.org>
  *  Copyright (c) 2002-2006 by Judd Vinet <jvinet@zeroflux.org>
  *  Copyright (c) 2005 by Aurelien Foret <orelien@chez.com>
  *  Copyright (c) 2005, 2006 by Christian Hamar <krics@linuxforum.hu>
@@ -21,8 +21,6 @@
  *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
  */
 
-#include "config.h"
-
 #include <stdlib.h>
 #include <string.h>
 #include <sys/types.h>
@@ -48,7 +46,7 @@ int SYMEXPORT alpm_pkg_free(alpm_pkg_t *pkg)
 	ASSERT(pkg != NULL, return -1);
 
 	/* Only free packages loaded in user space */
-	if(pkg->origin == PKG_FROM_FILE) {
+	if(pkg->origin == ALPM_PKG_FROM_FILE) {
 		_alpm_pkg_free(pkg);
 	}
 
@@ -64,12 +62,12 @@ int SYMEXPORT alpm_pkg_checkmd5sum(alpm_pkg_t *pkg)
 	ASSERT(pkg != NULL, return -1);
 	pkg->handle->pm_errno = 0;
 	/* We only inspect packages from sync repositories */
-	ASSERT(pkg->origin == PKG_FROM_SYNCDB,
+	ASSERT(pkg->origin == ALPM_PKG_FROM_SYNCDB,
 			RET_ERR(pkg->handle, ALPM_ERR_WRONG_ARGS, -1));
 
 	fpath = _alpm_filecache_find(pkg->handle, pkg->filename);
 
-	retval = _alpm_test_checksum(fpath, pkg->md5sum, ALPM_CSUM_MD5);
+	retval = _alpm_test_checksum(fpath, pkg->md5sum, ALPM_PKG_VALIDATION_MD5SUM);
 
 	if(retval == 0) {
 		return 0;
@@ -87,12 +85,13 @@ int SYMEXPORT alpm_pkg_checkmd5sum(alpm_pkg_t *pkg)
  * populated package structures. */
 static const char *_pkg_get_desc(alpm_pkg_t *pkg)        { return pkg->desc; }
 static const char *_pkg_get_url(alpm_pkg_t *pkg)         { return pkg->url; }
-static time_t _pkg_get_builddate(alpm_pkg_t *pkg)        { return pkg->builddate; }
-static time_t _pkg_get_installdate(alpm_pkg_t *pkg)      { return pkg->installdate; }
+static alpm_time_t _pkg_get_builddate(alpm_pkg_t *pkg)   { return pkg->builddate; }
+static alpm_time_t _pkg_get_installdate(alpm_pkg_t *pkg) { return pkg->installdate; }
 static const char *_pkg_get_packager(alpm_pkg_t *pkg)    { return pkg->packager; }
 static const char *_pkg_get_arch(alpm_pkg_t *pkg)        { return pkg->arch; }
 static off_t _pkg_get_isize(alpm_pkg_t *pkg)             { return pkg->isize; }
-static alpm_pkgreason_t _pkg_get_reason(alpm_pkg_t *pkg)    { return pkg->reason; }
+static alpm_pkgreason_t _pkg_get_reason(alpm_pkg_t *pkg) { return pkg->reason; }
+static alpm_pkgvalidation_t _pkg_get_validation(alpm_pkg_t *pkg) { return pkg->validation; }
 static int _pkg_has_scriptlet(alpm_pkg_t *pkg)           { return pkg->scriptlet; }
 
 static alpm_list_t *_pkg_get_licenses(alpm_pkg_t *pkg)   { return pkg->licenses; }
@@ -122,6 +121,23 @@ static int _pkg_changelog_close(const alpm_pkg_t UNUSED *pkg,
 	return EOF;
 }
 
+static struct archive *_pkg_mtree_open(alpm_pkg_t UNUSED *pkg)
+{
+	return NULL;
+}
+
+static int _pkg_mtree_next(const alpm_pkg_t UNUSED *pkg,
+		struct archive UNUSED *archive, struct archive_entry UNUSED **entry)
+{
+	return -1;
+}
+
+static int _pkg_mtree_close(const alpm_pkg_t UNUSED *pkg,
+		struct archive UNUSED *archive)
+{
+	return -1;
+}
+
 static int _pkg_force_load(alpm_pkg_t UNUSED *pkg) { return 0; }
 
 /** The standard package operations struct. Get fields directly from the
@@ -136,6 +152,7 @@ struct pkg_operations default_pkg_ops = {
 	.get_arch        = _pkg_get_arch,
 	.get_isize       = _pkg_get_isize,
 	.get_reason      = _pkg_get_reason,
+	.get_validation  = _pkg_get_validation,
 	.has_scriptlet   = _pkg_has_scriptlet,
 
 	.get_licenses    = _pkg_get_licenses,
@@ -152,6 +169,10 @@ struct pkg_operations default_pkg_ops = {
 	.changelog_read  = _pkg_changelog_read,
 	.changelog_close = _pkg_changelog_close,
 
+	.mtree_open      = _pkg_mtree_open,
+	.mtree_next      = _pkg_mtree_next,
+	.mtree_close     = _pkg_mtree_close,
+
 	.force_load      = _pkg_force_load,
 };
 
@@ -200,14 +221,14 @@ const char SYMEXPORT *alpm_pkg_get_url(alpm_pkg_t *pkg)
 	return pkg->ops->get_url(pkg);
 }
 
-time_t SYMEXPORT alpm_pkg_get_builddate(alpm_pkg_t *pkg)
+alpm_time_t SYMEXPORT alpm_pkg_get_builddate(alpm_pkg_t *pkg)
 {
 	ASSERT(pkg != NULL, return -1);
 	pkg->handle->pm_errno = 0;
 	return pkg->ops->get_builddate(pkg);
 }
 
-time_t SYMEXPORT alpm_pkg_get_installdate(alpm_pkg_t *pkg)
+alpm_time_t SYMEXPORT alpm_pkg_get_installdate(alpm_pkg_t *pkg)
 {
 	ASSERT(pkg != NULL, return -1);
 	pkg->handle->pm_errno = 0;
@@ -270,6 +291,13 @@ alpm_pkgreason_t SYMEXPORT alpm_pkg_get_reason(alpm_pkg_t *pkg)
 	return pkg->ops->get_reason(pkg);
 }
 
+alpm_pkgvalidation_t SYMEXPORT alpm_pkg_get_validation(alpm_pkg_t *pkg)
+{
+	ASSERT(pkg != NULL, return -1);
+	pkg->handle->pm_errno = 0;
+	return pkg->ops->get_validation(pkg);
+}
+
 alpm_list_t SYMEXPORT *alpm_pkg_get_licenses(alpm_pkg_t *pkg)
 {
 	ASSERT(pkg != NULL, return NULL);
@@ -344,7 +372,7 @@ alpm_db_t SYMEXPORT *alpm_pkg_get_db(alpm_pkg_t *pkg)
 {
 	/* Sanity checks */
 	ASSERT(pkg != NULL, return NULL);
-	ASSERT(pkg->origin != PKG_FROM_FILE, return NULL);
+	ASSERT(pkg->origin != ALPM_PKG_FROM_FILE, return NULL);
 	pkg->handle->pm_errno = 0;
 
 	return pkg->origin_data.db;
@@ -375,6 +403,31 @@ int SYMEXPORT alpm_pkg_changelog_close(const alpm_pkg_t *pkg, void *fp)
 	return pkg->ops->changelog_close(pkg, fp);
 }
 
+/** Open a package mtree file for reading. */
+struct archive SYMEXPORT *alpm_pkg_mtree_open(alpm_pkg_t * pkg)
+{
+	ASSERT(pkg != NULL, return NULL);
+	pkg->handle->pm_errno = 0;
+	return pkg->ops->mtree_open(pkg);
+}
+
+/** Read entry from an open mtree file. */
+int SYMEXPORT alpm_pkg_mtree_next(const alpm_pkg_t * pkg, struct archive *archive,
+	struct archive_entry **entry)
+{
+	ASSERT(pkg != NULL, return -1);
+	pkg->handle->pm_errno = 0;
+	return pkg->ops->mtree_next(pkg, archive, entry);
+}
+
+/** Close a package mtree file for reading. */
+int SYMEXPORT alpm_pkg_mtree_close(const alpm_pkg_t * pkg, struct archive *archive)
+{
+	ASSERT(pkg != NULL, return -1);
+	pkg->handle->pm_errno = 0;
+	return pkg->ops->mtree_close(pkg, archive);
+}
+
 int SYMEXPORT alpm_pkg_has_scriptlet(alpm_pkg_t *pkg)
 {
 	ASSERT(pkg != NULL, return -1);
@@ -382,7 +435,8 @@ int SYMEXPORT alpm_pkg_has_scriptlet(alpm_pkg_t *pkg)
 	return pkg->ops->has_scriptlet(pkg);
 }
 
-static void find_requiredby(alpm_pkg_t *pkg, alpm_db_t *db, alpm_list_t **reqs)
+static void find_requiredby(alpm_pkg_t *pkg, alpm_db_t *db, alpm_list_t **reqs,
+		int optional)
 {
 	const alpm_list_t *i;
 	pkg->handle->pm_errno = 0;
@@ -390,7 +444,14 @@ static void find_requiredby(alpm_pkg_t *pkg, alpm_db_t *db, alpm_list_t **reqs)
 	for(i = _alpm_db_get_pkgcache(db); i; i = i->next) {
 		alpm_pkg_t *cachepkg = i->data;
 		alpm_list_t *j;
-		for(j = alpm_pkg_get_depends(cachepkg); j; j = j->next) {
+
+		if(optional == 0) {
+			j = alpm_pkg_get_depends(cachepkg);
+		} else {
+			j = alpm_pkg_get_optdepends(cachepkg);
+		}
+
+		for(; j; j = j->next) {
 			if(_alpm_depcmp(pkg, j->data)) {
 				const char *cachepkgname = cachepkg->name;
 				if(alpm_list_find_str(*reqs, cachepkgname) == NULL) {
@@ -401,8 +462,7 @@ static void find_requiredby(alpm_pkg_t *pkg, alpm_db_t *db, alpm_list_t **reqs)
 	}
 }
 
-/** Compute the packages requiring a given package. */
-alpm_list_t SYMEXPORT *alpm_pkg_compute_requiredby(alpm_pkg_t *pkg)
+static alpm_list_t *compute_requiredby(alpm_pkg_t *pkg, int optional)
 {
 	const alpm_list_t *i;
 	alpm_list_t *reqs = NULL;
@@ -411,19 +471,19 @@ alpm_list_t SYMEXPORT *alpm_pkg_compute_requiredby(alpm_pkg_t *pkg)
 	ASSERT(pkg != NULL, return NULL);
 	pkg->handle->pm_errno = 0;
 
-	if(pkg->origin == PKG_FROM_FILE) {
+	if(pkg->origin == ALPM_PKG_FROM_FILE) {
 		/* The sane option; search locally for things that require this. */
-		find_requiredby(pkg, pkg->handle->db_local, &reqs);
+		find_requiredby(pkg, pkg->handle->db_local, &reqs, optional);
 	} else {
 		/* We have a DB package. if it is a local package, then we should
 		 * only search the local DB; else search all known sync databases. */
 		db = pkg->origin_data.db;
 		if(db->status & DB_STATUS_LOCAL) {
-			find_requiredby(pkg, db, &reqs);
+			find_requiredby(pkg, db, &reqs, optional);
 		} else {
 			for(i = pkg->handle->dbs_sync; i; i = i->next) {
 				db = i->data;
-				find_requiredby(pkg, db, &reqs);
+				find_requiredby(pkg, db, &reqs, optional);
 			}
 			reqs = alpm_list_msort(reqs, alpm_list_count(reqs), _alpm_str_cmp);
 		}
@@ -431,6 +491,19 @@ alpm_list_t SYMEXPORT *alpm_pkg_compute_requiredby(alpm_pkg_t *pkg)
 	return reqs;
 }
 
+/** Compute the packages requiring a given package. */
+alpm_list_t SYMEXPORT *alpm_pkg_compute_requiredby(alpm_pkg_t *pkg)
+{
+	return compute_requiredby(pkg, 0);
+}
+
+/** Compute the packages optionally requiring a given package. */
+alpm_list_t SYMEXPORT *alpm_pkg_compute_optionalfor(alpm_pkg_t *pkg)
+{
+	return compute_requiredby(pkg, 1);
+}
+
+
 /** @} */
 
 alpm_file_t *_alpm_file_copy(alpm_file_t *dest,
@@ -443,24 +516,24 @@ alpm_file_t *_alpm_file_copy(alpm_file_t *dest,
 	return dest;
 }
 
-/* Helper function for comparing files list entries
- */
-int _alpm_files_cmp(const void *f1, const void *f2)
-{
-	const alpm_file_t *file1 = f1;
-	const alpm_file_t *file2 = f2;
-	return strcmp(file1->name, file2->name);
-}
-
 alpm_pkg_t *_alpm_pkg_new(void)
 {
-	alpm_pkg_t* pkg;
+	alpm_pkg_t *pkg;
 
 	CALLOC(pkg, 1, sizeof(alpm_pkg_t), return NULL);
 
 	return pkg;
 }
 
+static alpm_list_t *list_depdup(alpm_list_t *old)
+{
+	alpm_list_t *i, *new = NULL;
+	for(i = old; i; i = i->next) {
+		new = alpm_list_add(new, _alpm_dep_dup(i->data));
+	}
+	return new;
+}
+
 /**
  * Duplicate a package data struct.
  * @param pkg the package to duplicate
@@ -509,10 +582,19 @@ int _alpm_pkg_dup(alpm_pkg_t *pkg, alpm_pkg_t **new_ptr)
 	newpkg->reason = pkg->reason;
 
 	newpkg->licenses   = alpm_list_strdup(pkg->licenses);
-	for(i = pkg->replaces; i; i = i->next) {
-		newpkg->replaces = alpm_list_add(newpkg->replaces, _alpm_dep_dup(i->data));
-	}
+	newpkg->replaces   = list_depdup(pkg->replaces);
 	newpkg->groups     = alpm_list_strdup(pkg->groups);
+	for(i = pkg->backup; i; i = i->next) {
+		newpkg->backup = alpm_list_add(newpkg->backup, _alpm_backup_dup(i->data));
+	}
+	newpkg->depends    = list_depdup(pkg->depends);
+	newpkg->optdepends = list_depdup(pkg->optdepends);
+	newpkg->conflicts  = list_depdup(pkg->conflicts);
+	newpkg->provides   = list_depdup(pkg->provides);
+	for(i = pkg->deltas; i; i = i->next) {
+		newpkg->deltas = alpm_list_add(newpkg->deltas, _alpm_delta_dup(i->data));
+	}
+
 	if(pkg->files.count) {
 		size_t filenum;
 		size_t len = sizeof(alpm_file_t) * pkg->files.count;
@@ -524,28 +606,15 @@ int _alpm_pkg_dup(alpm_pkg_t *pkg, alpm_pkg_t **new_ptr)
 			}
 		}
 		newpkg->files.count = pkg->files.count;
-	}
-	for(i = pkg->backup; i; i = i->next) {
-		newpkg->backup = alpm_list_add(newpkg->backup, _alpm_backup_dup(i->data));
-	}
-	for(i = pkg->depends; i; i = i->next) {
-		newpkg->depends = alpm_list_add(newpkg->depends, _alpm_dep_dup(i->data));
-	}
-	newpkg->optdepends = alpm_list_strdup(pkg->optdepends);
-	for(i = pkg->conflicts; i; i = i->next) {
-		newpkg->conflicts = alpm_list_add(newpkg->conflicts, _alpm_dep_dup(i->data));
-	}
-	for(i = pkg->provides; i; i = i->next) {
-		newpkg->provides = alpm_list_add(newpkg->provides, _alpm_dep_dup(i->data));
-	}
-	for(i = pkg->deltas; i; i = i->next) {
-		newpkg->deltas = alpm_list_add(newpkg->deltas, _alpm_delta_dup(i->data));
+		/* deliberately do not copy resolved_path as this is only used
+		 * during conflict checking and the sorting of list does not readily
+		 * allow keeping its efficient memory usage when copying */
 	}
 
 	/* internal */
 	newpkg->infolevel = pkg->infolevel;
 	newpkg->origin = pkg->origin;
-	if(newpkg->origin == PKG_FROM_FILE) {
+	if(newpkg->origin == ALPM_PKG_FROM_FILE) {
 		newpkg->origin_data.file = strdup(pkg->origin_data.file);
 	} else {
 		newpkg->origin_data.db = pkg->origin_data.db;
@@ -561,6 +630,12 @@ cleanup:
 	RET_ERR(pkg->handle, ALPM_ERR_MEMORY, -1);
 }
 
+static void free_deplist(alpm_list_t *deps)
+{
+	alpm_list_free_inner(deps, (alpm_list_fn_free)_alpm_dep_free);
+	alpm_list_free(deps);
+}
+
 void _alpm_pkg_free(alpm_pkg_t *pkg)
 {
 	if(pkg == NULL) {
@@ -579,39 +654,48 @@ void _alpm_pkg_free(alpm_pkg_t *pkg)
 	FREE(pkg->arch);
 
 	FREELIST(pkg->licenses);
-	alpm_list_free_inner(pkg->replaces, (alpm_list_fn_free)_alpm_dep_free);
-	alpm_list_free(pkg->replaces);
+	free_deplist(pkg->replaces);
 	FREELIST(pkg->groups);
 	if(pkg->files.count) {
-		size_t i;
-		for(i = 0; i < pkg->files.count; i++) {
-			free(pkg->files.files[i].name);
+		size_t i, j, k;
+		if(pkg->files.resolved_path) {
+			for(i = 0, j = 0; i < pkg->files.count; i++) {
+				for(k = j; k <= pkg->files.count; k++) {
+					if(pkg->files.resolved_path[i] == pkg->files.files[k].name) {
+						pkg->files.files[k].name = NULL;
+						j = k + 1;
+						break;
+					}
+				}
+				free(pkg->files.resolved_path[i]);
+			}
+			free(pkg->files.resolved_path);
+		}
+		for(j = 0; j < pkg->files.count; j++) {
+			FREE(pkg->files.files[j].name);
 		}
 		free(pkg->files.files);
 	}
 	alpm_list_free_inner(pkg->backup, (alpm_list_fn_free)_alpm_backup_free);
 	alpm_list_free(pkg->backup);
-	alpm_list_free_inner(pkg->depends, (alpm_list_fn_free)_alpm_dep_free);
-	alpm_list_free(pkg->depends);
-	FREELIST(pkg->optdepends);
-	alpm_list_free_inner(pkg->conflicts, (alpm_list_fn_free)_alpm_dep_free);
-	alpm_list_free(pkg->conflicts);
-	alpm_list_free_inner(pkg->provides, (alpm_list_fn_free)_alpm_dep_free);
-	alpm_list_free(pkg->provides);
+	free_deplist(pkg->depends);
+	free_deplist(pkg->optdepends);
+	free_deplist(pkg->conflicts);
+	free_deplist(pkg->provides);
 	alpm_list_free_inner(pkg->deltas, (alpm_list_fn_free)_alpm_delta_free);
 	alpm_list_free(pkg->deltas);
 	alpm_list_free(pkg->delta_path);
 	alpm_list_free(pkg->removes);
 
-	if(pkg->origin == PKG_FROM_FILE) {
+	if(pkg->origin == ALPM_PKG_FROM_FILE) {
 		FREE(pkg->origin_data.file);
 	}
 	FREE(pkg);
 }
 
 /* This function should be used when removing a target from upgrade/sync target list
- * Case 1: If pkg is a loaded package file (PKG_FROM_FILE), it will be freed.
- * Case 2: If pkg is a pkgcache entry (PKG_FROM_CACHE), it won't be freed,
+ * Case 1: If pkg is a loaded package file (ALPM_PKG_FROM_FILE), it will be freed.
+ * Case 2: If pkg is a pkgcache entry (ALPM_PKG_FROM_CACHE), it won't be freed,
  *         only the transaction specific fields of pkg will be freed.
  */
 void _alpm_pkg_free_trans(alpm_pkg_t *pkg)
@@ -620,7 +704,7 @@ void _alpm_pkg_free_trans(alpm_pkg_t *pkg)
 		return;
 	}
 
-	if(pkg->origin == PKG_FROM_FILE) {
+	if(pkg->origin == ALPM_PKG_FROM_FILE) {
 		_alpm_pkg_free(pkg);
 		return;
 	}
@@ -647,7 +731,7 @@ int _alpm_pkg_cmp(const void *p1, const void *p2)
 /* Test for existence of a package in a alpm_list_t*
  * of alpm_pkg_t*
  */
-alpm_pkg_t *_alpm_pkg_find(alpm_list_t *haystack, const char *needle)
+alpm_pkg_t SYMEXPORT *alpm_pkg_find(alpm_list_t *haystack, const char *needle)
 {
 	alpm_list_t *lp;
 	unsigned long needle_hash;
@@ -690,14 +774,14 @@ int _alpm_pkg_should_ignore(alpm_handle_t *handle, alpm_pkg_t *pkg)
 	alpm_list_t *groups = NULL;
 
 	/* first see if the package is ignored */
-	if(alpm_list_find_str(handle->ignorepkg, pkg->name)) {
+	if(alpm_list_find(handle->ignorepkg, pkg->name, _alpm_fnmatch)) {
 		return 1;
 	}
 
 	/* next see if the package is in a group that is ignored */
-	for(groups = handle->ignoregroup; groups; groups = groups->next) {
+	for(groups = alpm_pkg_get_groups(pkg); groups; groups = groups->next) {
 		char *grp = groups->data;
-		if(alpm_list_find_str(alpm_pkg_get_groups(pkg), grp)) {
+		if(alpm_list_find(handle->ignoregroup, grp, _alpm_fnmatch)) {
 			return 1;
 		}
 	}
