@@ -1,7 +1,7 @@
 /*
  *  package.h
  *
- *  Copyright (c) 2006-2011 Pacman Development Team <pacman-dev@archlinux.org>
+ *  Copyright (c) 2006-2013 Pacman Development Team <pacman-dev@archlinux.org>
  *  Copyright (c) 2002-2006 by Judd Vinet <jvinet@zeroflux.org>
  *  Copyright (c) 2005 by Aurelien Foret <orelien@chez.com>
  *  Copyright (c) 2006 by David Kimpe <dnaku@frugalware.org>
@@ -24,10 +24,11 @@
 #ifndef _ALPM_PACKAGE_H
 #define _ALPM_PACKAGE_H
 
-#include "config.h" /* ensure off_t is correct length */
-
 #include <sys/types.h> /* off_t */
-#include <time.h> /* time_t */
+
+/* libarchive */
+#include <archive.h>
+#include <archive_entry.h>
 
 #include "alpm.h"
 #include "backup.h"
@@ -44,12 +45,13 @@
 struct pkg_operations {
 	const char *(*get_desc) (alpm_pkg_t *);
 	const char *(*get_url) (alpm_pkg_t *);
-	time_t (*get_builddate) (alpm_pkg_t *);
-	time_t (*get_installdate) (alpm_pkg_t *);
+	alpm_time_t (*get_builddate) (alpm_pkg_t *);
+	alpm_time_t (*get_installdate) (alpm_pkg_t *);
 	const char *(*get_packager) (alpm_pkg_t *);
 	const char *(*get_arch) (alpm_pkg_t *);
 	off_t (*get_isize) (alpm_pkg_t *);
 	alpm_pkgreason_t (*get_reason) (alpm_pkg_t *);
+	alpm_pkgvalidation_t (*get_validation) (alpm_pkg_t *);
 	int (*has_scriptlet) (alpm_pkg_t *);
 
 	alpm_list_t *(*get_licenses) (alpm_pkg_t *);
@@ -66,6 +68,10 @@ struct pkg_operations {
 	size_t (*changelog_read) (void *, size_t, const alpm_pkg_t *, void *);
 	int (*changelog_close) (const alpm_pkg_t *, void *);
 
+	struct archive *(*mtree_open) (alpm_pkg_t *);
+	int (*mtree_next) (const alpm_pkg_t *, struct archive *, struct archive_entry **);
+	int (*mtree_close) (const alpm_pkg_t *, struct archive *);
+
 	int (*force_load) (alpm_pkg_t *);
 };
 
@@ -89,8 +95,8 @@ struct __alpm_pkg_t {
 	char *base64_sig;
 	char *arch;
 
-	time_t builddate;
-	time_t installdate;
+	alpm_time_t builddate;
+	alpm_time_t installdate;
 
 	off_t size;
 	off_t isize;
@@ -99,6 +105,7 @@ struct __alpm_pkg_t {
 	int scriptlet;
 
 	alpm_pkgreason_t reason;
+	alpm_pkgvalidation_t validation;
 	alpm_dbinfrq_t infolevel;
 	alpm_pkgfrom_t origin;
 	/* origin == PKG_FROM_FILE, use pkg->origin_data.file
@@ -127,22 +134,20 @@ struct __alpm_pkg_t {
 };
 
 alpm_file_t *_alpm_file_copy(alpm_file_t *dest, const alpm_file_t *src);
-int _alpm_files_cmp(const void *f1, const void *f2);
 
-alpm_pkg_t* _alpm_pkg_new(void);
+alpm_pkg_t *_alpm_pkg_new(void);
 int _alpm_pkg_dup(alpm_pkg_t *pkg, alpm_pkg_t **new_ptr);
 void _alpm_pkg_free(alpm_pkg_t *pkg);
 void _alpm_pkg_free_trans(alpm_pkg_t *pkg);
 
 int _alpm_pkg_validate_internal(alpm_handle_t *handle,
 		const char *pkgfile, alpm_pkg_t *syncpkg, alpm_siglevel_t level,
-		alpm_siglist_t **sigdata);
+		alpm_siglist_t **sigdata, alpm_pkgvalidation_t *validation);
 alpm_pkg_t *_alpm_pkg_load_internal(alpm_handle_t *handle,
 		const char *pkgfile, int full);
 
 int _alpm_pkg_cmp(const void *p1, const void *p2);
 int _alpm_pkg_compare_versions(alpm_pkg_t *local_pkg, alpm_pkg_t *pkg);
-alpm_pkg_t *_alpm_pkg_find(alpm_list_t *haystack, const char *needle);
 int _alpm_pkg_should_ignore(alpm_handle_t *handle, alpm_pkg_t *pkg);
 
 #endif /* _ALPM_PACKAGE_H */
