@@ -1,7 +1,7 @@
 /*
  *  callback.c
  *
- *  Copyright (c) 2006-2011 Pacman Development Team <pacman-dev@archlinux.org>
+ *  Copyright (c) 2006-2013 Pacman Development Team <pacman-dev@archlinux.org>
  *  Copyright (c) 2002-2006 by Judd Vinet <jvinet@zeroflux.org>
  *
  *  This program is free software; you can redistribute it and/or modify
@@ -18,13 +18,12 @@
  *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
  */
 
-#include "config.h"
-
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 #include <sys/time.h>
 #include <sys/types.h> /* off_t */
+#include <time.h>
 #include <unistd.h>
 #include <wchar.h>
 #include <limits.h> /* UINT_MAX */
@@ -32,6 +31,7 @@
 #include <alpm.h>
 
 /* pacman */
+#include "pacman.h"
 #include "callback.h"
 #include "util.h"
 #include "conf.h"
@@ -44,15 +44,19 @@ static off_t list_total = 0.0;
 static int on_progress = 0;
 static alpm_list_t *output = NULL;
 
-/* Silly little helper function, determines if the caller needs a visual update
+/* update speed for the fill_progress based functions */
+#define UPDATE_SPEED_MS 200
+
+/**
+ * Silly little helper function, determines if the caller needs a visual update
  * since the last time this function was called.
- * This is made for the two progress bar functions, to prevent flicker
- *
- * first_call indicates if this is the first time it is called, for
- * initialization purposes */
-static double get_update_timediff(int first_call)
+ * This is made for the two progress bar functions, to prevent flicker.
+ * @param first_call 1 on first call for initialization purposes, 0 otherwise
+ * @return number of milliseconds since last call
+ */
+static long get_update_timediff(int first_call)
 {
-	double retval = 0.0;
+	long retval = 0;
 	static struct timeval last_time = {0, 0};
 
 	/* on first call, simply set the last time and return */
@@ -60,18 +64,17 @@ static double get_update_timediff(int first_call)
 		gettimeofday(&last_time, NULL);
 	} else {
 		struct timeval this_time;
-		double diff_sec, diff_usec;
+		time_t diff_sec;
+		suseconds_t diff_usec;
 
 		gettimeofday(&this_time, NULL);
 		diff_sec = this_time.tv_sec - last_time.tv_sec;
 		diff_usec = this_time.tv_usec - last_time.tv_usec;
 
-		retval = diff_sec + (diff_usec / 1000000.0);
+		retval = (diff_sec * 1000) + (diff_usec / 1000);
 
-		/* return 0 and do not update last_time if interval was too short */
-		if(retval < UPDATE_SPEED_SEC) {
-			retval = 0.0;
-		} else {
+		/* do not update last_time if interval was too short */
+		if(retval >= UPDATE_SPEED_MS) {
 			last_time = this_time;
 		}
 	}
@@ -95,41 +98,41 @@ static void fill_progress(const int bar_percent, const int disp_percent,
 	}
 
 	if(hashlen > 0) {
-		printf(" [");
+		fputs(" [", stdout);
 		for(i = hashlen; i > 0; --i) {
 			/* if special progress bar enabled */
 			if(config->chomp) {
 				if(i > hashlen - hash) {
-					printf("-");
+					putchar('-');
 				} else if(i == hashlen - hash) {
 					if(lasthash == hash) {
 						if(mouth) {
-							printf("\033[1;33mC\033[m");
+							fputs("\033[1;33mC\033[m", stdout);
 						} else {
-							printf("\033[1;33mc\033[m");
+							fputs("\033[1;33mc\033[m", stdout);
 						}
 					} else {
 						lasthash = hash;
 						mouth = mouth == 1 ? 0 : 1;
 						if(mouth) {
-							printf("\033[1;33mC\033[m");
+							fputs("\033[1;33mC\033[m", stdout);
 						} else {
-							printf("\033[1;33mc\033[m");
+							fputs("\033[1;33mc\033[m", stdout);
 						}
 					}
-				} else if(i%3 == 0) {
-					printf("\033[0;37mo\033[m");
+				} else if(i % 3 == 0) {
+					fputs("\033[0;37mo\033[m", stdout);
 				} else {
-					printf("\033[0;37m \033[m");
+					fputs("\033[0;37m \033[m", stdout);
 				}
 			} /* else regular progress bar */
 			else if(i > hashlen - hash) {
-				printf("#");
+				putchar('#');
 			} else {
-				printf("-");
+				putchar('-');
 			}
 		}
-		printf("]");
+		putchar(']');
 	}
 	/* print display percent after progress bar */
 	/* 5 = 1 space + 3 digits + 1 % */
@@ -138,9 +141,9 @@ static void fill_progress(const int bar_percent, const int disp_percent,
 	}
 
 	if(bar_percent == 100) {
-		printf("\n");
+		putchar('\n');
 	} else {
-		printf("\r");
+		putchar('\r');
 	}
 	fflush(stdout);
 }
@@ -174,9 +177,10 @@ void cb_event(alpm_event_t event, void *data1, void *data2)
 			}
 			break;
 		case ALPM_EVENT_ADD_DONE:
-			alpm_logaction(config->handle, "installed %s (%s)\n",
-			         alpm_pkg_get_name(data1),
-			         alpm_pkg_get_version(data1));
+			alpm_logaction(config->handle, PACMAN_CALLER_PREFIX,
+					"installed %s (%s)\n",
+					alpm_pkg_get_name(data1),
+					alpm_pkg_get_version(data1));
 			display_optdepends(data1);
 			break;
 		case ALPM_EVENT_REMOVE_START:
@@ -185,9 +189,10 @@ void cb_event(alpm_event_t event, void *data1, void *data2)
 			}
 			break;
 		case ALPM_EVENT_REMOVE_DONE:
-			alpm_logaction(config->handle, "removed %s (%s)\n",
-			         alpm_pkg_get_name(data1),
-			         alpm_pkg_get_version(data1));
+			alpm_logaction(config->handle, PACMAN_CALLER_PREFIX,
+					"removed %s (%s)\n",
+					alpm_pkg_get_name(data1),
+					alpm_pkg_get_version(data1));
 			break;
 		case ALPM_EVENT_UPGRADE_START:
 			if(config->noprogressbar) {
@@ -195,17 +200,50 @@ void cb_event(alpm_event_t event, void *data1, void *data2)
 			}
 			break;
 		case ALPM_EVENT_UPGRADE_DONE:
-			alpm_logaction(config->handle, "upgraded %s (%s -> %s)\n",
-			         (char *)alpm_pkg_get_name(data1),
-			         (char *)alpm_pkg_get_version(data2),
-			         (char *)alpm_pkg_get_version(data1));
-			display_new_optdepends(data2,data1);
+			alpm_logaction(config->handle, PACMAN_CALLER_PREFIX,
+					"upgraded %s (%s -> %s)\n",
+					alpm_pkg_get_name(data1),
+					alpm_pkg_get_version(data2),
+					alpm_pkg_get_version(data1));
+			display_new_optdepends(data2, data1);
+			break;
+		case ALPM_EVENT_DOWNGRADE_START:
+			if(config->noprogressbar) {
+				printf(_("downgrading %s...\n"), alpm_pkg_get_name(data1));
+			}
+			break;
+		case ALPM_EVENT_DOWNGRADE_DONE:
+			alpm_logaction(config->handle, PACMAN_CALLER_PREFIX,
+					"downgraded %s (%s -> %s)\n",
+					alpm_pkg_get_name(data1),
+					alpm_pkg_get_version(data2),
+					alpm_pkg_get_version(data1));
+			display_new_optdepends(data2, data1);
+			break;
+		case ALPM_EVENT_REINSTALL_START:
+			if(config->noprogressbar) {
+				printf(_("reinstalling %s...\n"), alpm_pkg_get_name(data1));
+			}
+			break;
+		case ALPM_EVENT_REINSTALL_DONE:
+			alpm_logaction(config->handle, PACMAN_CALLER_PREFIX,
+					"reinstalled %s (%s)\n",
+					alpm_pkg_get_name(data1),
+					alpm_pkg_get_version(data1));
 			break;
 		case ALPM_EVENT_INTEGRITY_START:
 			if(config->noprogressbar) {
 				printf(_("checking package integrity...\n"));
 			}
 			break;
+		case ALPM_EVENT_KEYRING_START:
+			if(config->noprogressbar) {
+				printf(_("checking keyring...\n"));
+			}
+			break;
+		case ALPM_EVENT_KEY_DOWNLOAD_START:
+			printf(_("downloading required keys...\n"));
+			break;
 		case ALPM_EVENT_LOAD_START:
 			if(config->noprogressbar) {
 				printf(_("loading package files...\n"));
@@ -227,22 +265,34 @@ void cb_event(alpm_event_t event, void *data1, void *data2)
 			printf(_("failed.\n"));
 			break;
 		case ALPM_EVENT_SCRIPTLET_INFO:
-			printf("%s", (char *)data1);
+			fputs((const char *)data1, stdout);
 			break;
 		case ALPM_EVENT_RETRIEVE_START:
-			printf(_(":: Retrieving packages from %s...\n"), (char *)data1);
+			colon_printf(_("Retrieving packages ...\n"));
 			break;
 		case ALPM_EVENT_DISKSPACE_START:
 			if(config->noprogressbar) {
 				printf(_("checking available disk space...\n"));
 			}
 			break;
+		case ALPM_EVENT_OPTDEP_REQUIRED:
+			colon_printf(_("%s optionally requires %s\n"), alpm_pkg_get_name(data1),
+				alpm_dep_compute_string(data2));
+			break;
+		case ALPM_EVENT_DATABASE_MISSING:
+			if(!config->op_s_sync) {
+				pm_printf(ALPM_LOG_WARNING,
+					"database file for '%s' does not exist\n", (char *)data1);
+			}
+			break;
 		/* all the simple done events, with fallthrough for each */
 		case ALPM_EVENT_FILECONFLICTS_DONE:
 		case ALPM_EVENT_CHECKDEPS_DONE:
 		case ALPM_EVENT_RESOLVEDEPS_DONE:
 		case ALPM_EVENT_INTERCONFLICTS_DONE:
 		case ALPM_EVENT_INTEGRITY_DONE:
+		case ALPM_EVENT_KEYRING_DONE:
+		case ALPM_EVENT_KEY_DOWNLOAD_DONE:
 		case ALPM_EVENT_LOAD_DONE:
 		case ALPM_EVENT_DELTA_INTEGRITY_DONE:
 		case ALPM_EVENT_DELTA_PATCHES_DONE:
@@ -264,14 +314,14 @@ void cb_question(alpm_question_t event, void *data1, void *data2,
 	switch(event) {
 		case ALPM_QUESTION_INSTALL_IGNOREPKG:
 			if(!config->op_s_downloadonly) {
-				*response = yesno(_(":: %s is in IgnorePkg/IgnoreGroup. Install anyway?"),
+				*response = yesno(_("%s is in IgnorePkg/IgnoreGroup. Install anyway?"),
 								  alpm_pkg_get_name(data1));
 			} else {
 				*response = 1;
 			}
 			break;
 		case ALPM_QUESTION_REPLACE_PKG:
-			*response = yesno(_(":: Replace %s with %s/%s?"),
+			*response = yesno(_("Replace %s with %s/%s?"),
 					alpm_pkg_get_name(data1),
 					(char *)data3,
 					alpm_pkg_get_name(data2));
@@ -280,12 +330,12 @@ void cb_question(alpm_question_t event, void *data1, void *data2,
 			/* data parameters: target package, local package, conflict (strings) */
 			/* print conflict only if it contains new information */
 			if(strcmp(data1, data3) == 0 || strcmp(data2, data3) == 0) {
-				*response = noyes(_(":: %s and %s are in conflict. Remove %s?"),
+				*response = noyes(_("%s and %s are in conflict. Remove %s?"),
 						(char *)data1,
 						(char *)data2,
 						(char *)data2);
 			} else {
-				*response = noyes(_(":: %s and %s are in conflict (%s). Remove %s?"),
+				*response = noyes(_("%s and %s are in conflict (%s). Remove %s?"),
 						(char *)data1,
 						(char *)data2,
 						(char *)data3,
@@ -294,19 +344,19 @@ void cb_question(alpm_question_t event, void *data1, void *data2,
 			break;
 		case ALPM_QUESTION_REMOVE_PKGS:
 			{
-				alpm_list_t *unresolved = (alpm_list_t *) data1;
+				alpm_list_t *unresolved = data1;
 				alpm_list_t *namelist = NULL, *i;
 				size_t count = 0;
-				for (i = unresolved; i; i = i->next) {
+				for(i = unresolved; i; i = i->next) {
 					namelist = alpm_list_add(namelist,
 							(char *)alpm_pkg_get_name(i->data));
 					count++;
 				}
-				printf(_n(
-							":: The following package cannot be upgraded due to unresolvable dependencies:\n",
-							":: The following packages cannot be upgraded due to unresolvable dependencies:\n",
+				colon_printf(_n(
+							"The following package cannot be upgraded due to unresolvable dependencies:\n",
+							"The following packages cannot be upgraded due to unresolvable dependencies:\n",
 							count));
-				list_display("     ", namelist);
+				list_display("     ", namelist, getcols(fileno(stdout)));
 				printf("\n");
 				*response = noyes(_n(
 							"Do you want to skip the above package for this upgrade?",
@@ -317,10 +367,10 @@ void cb_question(alpm_question_t event, void *data1, void *data2,
 			break;
 		case ALPM_QUESTION_SELECT_PROVIDER:
 			{
-				alpm_list_t *providers = (alpm_list_t *)data1;
+				alpm_list_t *providers = data1;
 				size_t count = alpm_list_count(providers);
 				char *depstring = alpm_dep_compute_string((alpm_depend_t *)data2);
-				printf(_(":: There are %zd providers available for %s:\n"), count,
+				colon_printf(_("There are %zd providers available for %s:\n"), count,
 						depstring);
 				free(depstring);
 				select_display(providers);
@@ -329,7 +379,7 @@ void cb_question(alpm_question_t event, void *data1, void *data2,
 			break;
 		case ALPM_QUESTION_LOCAL_NEWER:
 			if(!config->op_s_downloadonly) {
-				*response = yesno(_(":: %s-%s: local version is newer. Upgrade anyway?"),
+				*response = yesno(_("%s-%s: local version is newer. Upgrade anyway?"),
 						alpm_pkg_get_name(data1),
 						alpm_pkg_get_version(data1));
 			} else {
@@ -337,18 +387,25 @@ void cb_question(alpm_question_t event, void *data1, void *data2,
 			}
 			break;
 		case ALPM_QUESTION_CORRUPTED_PKG:
-			*response = yesno(_(":: File %s is corrupted (%s).\n"
+			*response = yesno(_("File %s is corrupted (%s).\n"
 						"Do you want to delete it?"),
 					(char *)data1,
-					alpm_strerror(*(enum _alpm_errno_t *)data2));
+					alpm_strerror(*(alpm_errno_t *)data2));
 			break;
 		case ALPM_QUESTION_IMPORT_KEY:
 			{
 				alpm_pgpkey_t *key = data1;
 				char created[12];
-				strftime(created, 12, "%Y-%m-%d", localtime(&(key->created)));
-				*response = yesno(_(":: Import PGP key %s, \"%s\", created %s?"),
-						key->fingerprint, key->uid, created);
+				const char *revoked = "";
+				time_t time = (time_t)key->created;
+				strftime(created, 12, "%Y-%m-%d", localtime(&time));
+
+				if(key->revoked) {
+					revoked = " (revoked)";
+				}
+
+				*response = yesno(_("Import PGP key %d%c/%s, \"%s\", created: %s%s?"),
+						key->length, key->pubkey_algo, key->fingerprint, key->uid, created, revoked);
 			}
 			break;
 	}
@@ -375,7 +432,7 @@ void cb_progress(alpm_progress_t event, const char *pkgname, int percent,
 	int len, wclen, wcwid, padwid;
 	wchar_t *wcstr;
 
-	const unsigned short cols = getcols();
+	const unsigned short cols = getcols(fileno(stdout));
 
 	if(config->noprogressbar || cols == 0) {
 		return;
@@ -393,7 +450,7 @@ void cb_progress(alpm_progress_t event, const char *pkgname, int percent,
 		if(current != prevcurrent) {
 			/* update always */
 		} else if(!pkgname || percent == prevpercent ||
-				get_update_timediff(0) < UPDATE_SPEED_SEC) {
+				get_update_timediff(0) < UPDATE_SPEED_MS) {
 			/* only update the progress bar when we have a package name, the
 			 * percentage has changed, and it has been long enough. */
 			return;
@@ -404,13 +461,19 @@ void cb_progress(alpm_progress_t event, const char *pkgname, int percent,
 	prevcurrent = current;
 
 	/* set text of message to display */
-	switch (event) {
+	switch(event) {
 		case ALPM_PROGRESS_ADD_START:
 			opr = _("installing");
 			break;
 		case ALPM_PROGRESS_UPGRADE_START:
 			opr = _("upgrading");
 			break;
+		case ALPM_PROGRESS_DOWNGRADE_START:
+			opr = _("downgrading");
+			break;
+		case ALPM_PROGRESS_REINSTALL_START:
+			opr = _("reinstalling");
+			break;
 		case ALPM_PROGRESS_REMOVE_START:
 			opr = _("removing");
 			break;
@@ -423,6 +486,9 @@ void cb_progress(alpm_progress_t event, const char *pkgname, int percent,
 		case ALPM_PROGRESS_INTEGRITY_START:
 			opr = _("checking package integrity");
 			break;
+		case ALPM_PROGRESS_KEYRING_START:
+			opr = _("checking keys in keyring");
+			break;
 		case ALPM_PROGRESS_LOAD_START:
 			opr = _("loading package files");
 			break;
@@ -493,7 +559,7 @@ void cb_progress(alpm_progress_t event, const char *pkgname, int percent,
 		alpm_list_t *i = NULL;
 		on_progress = 0;
 		for(i = output; i; i = i->next) {
-			printf("%s", (char *)i->data);
+			fputs((const char *)i->data, stdout);
 		}
 		fflush(stdout);
 		FREELIST(output);
@@ -528,13 +594,14 @@ void cb_dl_progress(const char *filename, off_t file_xfered, off_t file_total)
 
 	int totaldownload = 0;
 	off_t xfered, total;
-	double rate = 0.0, timediff = 0.0;
+	double rate = 0.0;
+	long timediff = 0;
 	unsigned int eta_h = 0, eta_m = 0, eta_s = 0;
 	double rate_human, xfered_human;
 	const char *rate_label, *xfered_label;
 	int file_percent = 0, total_percent = 0;
 
-	const unsigned short cols = getcols();
+	const unsigned short cols = getcols(fileno(stdout));
 
 	if(config->noprogressbar || cols == 0 || file_total == -1) {
 		if(file_xfered == 0) {
@@ -587,16 +654,17 @@ void cb_dl_progress(const char *filename, off_t file_xfered, off_t file_total)
 	} else if(file_xfered == file_total) {
 		/* compute final values */
 		struct timeval current_time;
-		double diff_sec, diff_usec;
+		time_t diff_sec;
+		suseconds_t diff_usec;
 
 		gettimeofday(&current_time, NULL);
 		diff_sec = current_time.tv_sec - initial_time.tv_sec;
 		diff_usec = current_time.tv_usec - initial_time.tv_usec;
-		timediff = diff_sec + (diff_usec / 1000000.0);
-		if(timediff > 0.0) {
-			rate = xfered / timediff;
-			/* round elapsed time to the nearest second */
-			eta_s = (unsigned int)(timediff + 0.5);
+		timediff = (diff_sec * 1000) + (diff_usec / 1000);
+		if(timediff > 0) {
+			rate = (double)xfered / (timediff / 1000.0);
+			/* round elapsed time (in ms) to the nearest second */
+			eta_s = (unsigned int)(timediff + 500) / 1000;
 		} else {
 			eta_s = 0;
 		}
@@ -604,11 +672,11 @@ void cb_dl_progress(const char *filename, off_t file_xfered, off_t file_total)
 		/* compute current average values */
 		timediff = get_update_timediff(0);
 
-		if(timediff < UPDATE_SPEED_SEC) {
+		if(timediff < UPDATE_SPEED_MS) {
 			/* return if the calling interval was too short */
 			return;
 		}
-		rate = (xfered - xfered_last) / timediff;
+		rate = (double)(xfered - xfered_last) / (timediff / 1000.0);
 		/* average rate to reduce jumpiness */
 		rate = (rate + 2 * rate_last) / 3;
 		if(rate > 0.0) {
@@ -713,7 +781,7 @@ void cb_dl_progress(const char *filename, off_t file_xfered, off_t file_total)
 	} else if(eta_h < 100) {
 		printf("%02u:%02u:%02u", eta_h, eta_m, eta_s);
 	} else {
-		printf("--:--");
+		fputs("--:--", stdout);
 	}
 
 	free(fname);
