@@ -1,7 +1,7 @@
 /*
  *  conf.c
  *
- *  Copyright (c) 2006-2011 Pacman Development Team <pacman-dev@archlinux.org>
+ *  Copyright (c) 2006-2013 Pacman Development Team <pacman-dev@archlinux.org>
  *  Copyright (c) 2002-2006 by Judd Vinet <jvinet@zeroflux.org>
  *
  *  This program is free software; you can redistribute it and/or modify
@@ -18,11 +18,10 @@
  *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
  */
 
-#include "config.h"
-
 #include <errno.h>
 #include <glob.h>
 #include <limits.h>
+#include <locale.h> /* setlocale */
 #include <fcntl.h> /* open */
 #include <stdlib.h>
 #include <stdio.h>
@@ -41,6 +40,45 @@
 /* global config variable */
 config_t *config = NULL;
 
+#define NOCOLOR       "\033[0m"
+
+#define BOLD          "\033[0;1m"
+
+#define BLACK         "\033[0;30m"
+#define RED           "\033[0;31m"
+#define GREEN         "\033[0;32m"
+#define YELLOW        "\033[0;33m"
+#define BLUE          "\033[0;34m"
+#define MAGENTA       "\033[0;35m"
+#define CYAN          "\033[0;36m"
+#define WHITE         "\033[0;37m"
+
+#define BOLDBLACK     "\033[1;30m"
+#define BOLDRED       "\033[1;31m"
+#define BOLDGREEN     "\033[1;32m"
+#define BOLDYELLOW    "\033[1;33m"
+#define BOLDBLUE      "\033[1;34m"
+#define BOLDMAGENTA   "\033[1;35m"
+#define BOLDCYAN      "\033[1;36m"
+#define BOLDWHITE     "\033[1;37m"
+
+void enable_colors(int colors)
+{
+	colstr_t *colstr = &config->colstr;
+
+	if(colors == PM_COLOR_ON) {
+		colstr->colon   = BOLDBLUE "::" BOLD " ";
+		colstr->title   = BOLD;
+		colstr->repo    = BOLDMAGENTA;
+		colstr->version = BOLDGREEN;
+		colstr->groups  = BOLDBLUE;
+		colstr->meta    = BOLDCYAN;
+		colstr->warn    = BOLDYELLOW;
+		colstr->err     = BOLDRED;
+		colstr->nocolor = NOCOLOR;
+	}
+}
+
 config_t *config_new(void)
 {
 	config_t *newconfig = calloc(1, sizeof(config_t));
@@ -54,11 +92,24 @@ config_t *config_new(void)
 	newconfig->op = PM_OP_MAIN;
 	newconfig->logmask = ALPM_LOG_ERROR | ALPM_LOG_WARNING;
 	newconfig->configfile = strdup(CONFFILE);
+	newconfig->deltaratio = 0.0;
 	if(alpm_capabilities() & ALPM_CAPABILITY_SIGNATURES) {
 		newconfig->siglevel = ALPM_SIG_PACKAGE | ALPM_SIG_PACKAGE_OPTIONAL |
 			ALPM_SIG_DATABASE | ALPM_SIG_DATABASE_OPTIONAL;
+		newconfig->localfilesiglevel = ALPM_SIG_USE_DEFAULT;
+		newconfig->remotefilesiglevel = ALPM_SIG_USE_DEFAULT;
 	}
 
+	newconfig->colstr.colon   = ":: ";
+	newconfig->colstr.title   = "";
+	newconfig->colstr.repo    = "";
+	newconfig->colstr.version = "";
+	newconfig->colstr.groups  = "";
+	newconfig->colstr.meta    = "";
+	newconfig->colstr.warn    = "";
+	newconfig->colstr.err     = "";
+	newconfig->colstr.nocolor = "";
+
 	return newconfig;
 }
 
@@ -72,7 +123,6 @@ int config_free(config_t *oldconfig)
 	alpm_list_free(oldconfig->explicit_removes);
 
 	FREELIST(oldconfig->holdpkg);
-	FREELIST(oldconfig->syncfirst);
 	FREELIST(oldconfig->ignorepkg);
 	FREELIST(oldconfig->ignoregrp);
 	FREELIST(oldconfig->noupgrade);
@@ -92,7 +142,8 @@ int config_free(config_t *oldconfig)
 }
 
 /** Helper function for download_with_xfercommand() */
-static char *get_filename(const char *url) {
+static char *get_filename(const char *url)
+{
 	char *filename = strrchr(url, '/');
 	if(filename != NULL) {
 		filename++;
@@ -101,7 +152,8 @@ static char *get_filename(const char *url) {
 }
 
 /** Helper function for download_with_xfercommand() */
-static char *get_destfile(const char *path, const char *filename) {
+static char *get_destfile(const char *path, const char *filename)
+{
 	char *destfile;
 	/* len = localpath len + filename len + null */
 	size_t len = strlen(path) + strlen(filename) + 1;
@@ -112,7 +164,8 @@ static char *get_destfile(const char *path, const char *filename) {
 }
 
 /** Helper function for download_with_xfercommand() */
-static char *get_tempfile(const char *path, const char *filename) {
+static char *get_tempfile(const char *path, const char *filename)
+{
 	char *tempfile;
 	/* len = localpath len + filename len + '.part' len + null */
 	size_t len = strlen(path) + strlen(filename) + 6;
@@ -124,12 +177,13 @@ static char *get_tempfile(const char *path, const char *filename) {
 
 /** External fetch callback */
 static int download_with_xfercommand(const char *url, const char *localpath,
-		int force) {
+		int force)
+{
 	int ret = 0, retval;
 	int usepart = 0;
 	int cwdfd;
 	struct stat st;
-	char *parsedcmd,*tempcmd;
+	char *parsedcmd, *tempcmd;
 	char *destfile, *tempfile, *filename;
 
 	if(!config->xfercommand) {
@@ -203,11 +257,14 @@ static int download_with_xfercommand(const char *url, const char *localpath,
 cleanup:
 	/* restore the old cwd if we have it */
 	if(cwdfd >= 0) {
+		int close_ret;
 		if(fchdir(cwdfd) != 0) {
 			pm_printf(ALPM_LOG_ERROR, _("could not restore working directory (%s)\n"),
 					strerror(errno));
 		}
-		close(cwdfd);
+		do {
+			close_ret = close(cwdfd);
+		} while(close_ret == -1 && errno == EINTR);
 	}
 
 	if(ret == -1) {
@@ -256,11 +313,11 @@ static int process_siglevel(alpm_list_t *values, alpm_siglevel_t *storage,
 		const char *original = i->data, *value;
 		int package = 0, database = 0;
 
-		if (strncmp(original, "Package", strlen("Package")) == 0) {
+		if(strncmp(original, "Package", strlen("Package")) == 0) {
 			/* only packages are affected, don't flip flags for databases */
 			value = original + strlen("Package");
 			package = 1;
-		} else if (strncmp(original, "Database", strlen("Database")) == 0) {
+		} else if(strncmp(original, "Database", strlen("Database")) == 0) {
 			/* only databases are affected, don't flip flags for packages */
 			value = original + strlen("Database");
 			database = 1;
@@ -274,6 +331,7 @@ static int process_siglevel(alpm_list_t *values, alpm_siglevel_t *storage,
 		if(strcmp(value, "Never") == 0) {
 			if(package) {
 				level &= ~ALPM_SIG_PACKAGE;
+				level |= ALPM_SIG_PACKAGE_SET;
 			}
 			if(database) {
 				level &= ~ALPM_SIG_DATABASE;
@@ -282,6 +340,7 @@ static int process_siglevel(alpm_list_t *values, alpm_siglevel_t *storage,
 			if(package) {
 				level |= ALPM_SIG_PACKAGE;
 				level |= ALPM_SIG_PACKAGE_OPTIONAL;
+				level |= ALPM_SIG_PACKAGE_SET;
 			}
 			if(database) {
 				level |= ALPM_SIG_DATABASE;
@@ -291,6 +350,7 @@ static int process_siglevel(alpm_list_t *values, alpm_siglevel_t *storage,
 			if(package) {
 				level |= ALPM_SIG_PACKAGE;
 				level &= ~ALPM_SIG_PACKAGE_OPTIONAL;
+				level |= ALPM_SIG_PACKAGE_SET;
 			}
 			if(database) {
 				level |= ALPM_SIG_DATABASE;
@@ -300,6 +360,7 @@ static int process_siglevel(alpm_list_t *values, alpm_siglevel_t *storage,
 			if(package) {
 				level &= ~ALPM_SIG_PACKAGE_MARGINAL_OK;
 				level &= ~ALPM_SIG_PACKAGE_UNKNOWN_OK;
+				level |= ALPM_SIG_PACKAGE_TRUST_SET;
 			}
 			if(database) {
 				level &= ~ALPM_SIG_DATABASE_MARGINAL_OK;
@@ -309,6 +370,7 @@ static int process_siglevel(alpm_list_t *values, alpm_siglevel_t *storage,
 			if(package) {
 				level |= ALPM_SIG_PACKAGE_MARGINAL_OK;
 				level |= ALPM_SIG_PACKAGE_UNKNOWN_OK;
+				level |= ALPM_SIG_PACKAGE_TRUST_SET;
 			}
 			if(database) {
 				level |= ALPM_SIG_DATABASE_MARGINAL_OK;
@@ -338,6 +400,30 @@ static int process_siglevel(alpm_list_t *values, alpm_siglevel_t *storage,
 	return ret;
 }
 
+/**
+ * Merge the package entires of two signature verification levels.
+ * @param base initial siglevel
+ * @param over overridden siglevel, derived value is stored here
+ */
+static void merge_siglevel(alpm_siglevel_t *base, alpm_siglevel_t *over)
+{
+	alpm_siglevel_t level = *over;
+	if(level & ALPM_SIG_USE_DEFAULT) {
+		level = *base;
+	} else {
+		if(!(level & ALPM_SIG_PACKAGE_SET)) {
+			level |= *base & ALPM_SIG_PACKAGE;
+			level |= *base & ALPM_SIG_PACKAGE_OPTIONAL;
+		}
+		if(!(level & ALPM_SIG_PACKAGE_TRUST_SET)) {
+			level |= *base & ALPM_SIG_PACKAGE_MARGINAL_OK;
+			level |= *base & ALPM_SIG_PACKAGE_UNKNOWN_OK;
+		}
+	}
+
+	*over = level;
+}
+
 static int process_cleanmethods(alpm_list_t *values,
 		const char *file, int linenum)
 {
@@ -368,17 +454,14 @@ static int process_cleanmethods(alpm_list_t *values,
 static void setrepeatingoption(char *ptr, const char *option,
 		alpm_list_t **list)
 {
-	char *q;
+	char *val, *saveptr;
 
-	while((q = strchr(ptr, ' '))) {
-		*q = '\0';
-		*list = alpm_list_add(*list, strdup(ptr));
-		pm_printf(ALPM_LOG_DEBUG, "config: %s: %s\n", option, ptr);
-		ptr = q;
-		ptr++;
+	val = strtok_r(ptr, " ", &saveptr);
+	while(val) {
+		*list = alpm_list_add(*list, strdup(val));
+		pm_printf(ALPM_LOG_DEBUG, "config: %s: %s\n", option, val);
+		val = strtok_r(NULL, " ", &saveptr);
 	}
-	*list = alpm_list_add(*list, strdup(ptr));
-	pm_printf(ALPM_LOG_DEBUG, "config: %s: %s\n", option, ptr);
 }
 
 static int _parse_options(const char *key, char *value,
@@ -396,13 +479,18 @@ static int _parse_options(const char *key, char *value,
 			config->verbosepkglists = 1;
 			pm_printf(ALPM_LOG_DEBUG, "config: verbosepkglists\n");
 		} else if(strcmp(key, "UseDelta") == 0) {
-			config->usedelta = 1;
-			pm_printf(ALPM_LOG_DEBUG, "config: usedelta\n");
+			config->deltaratio = 0.7;
+			pm_printf(ALPM_LOG_DEBUG, "config: usedelta (default 0.7)\n");
 		} else if(strcmp(key, "TotalDownload") == 0) {
 			config->totaldownload = 1;
 			pm_printf(ALPM_LOG_DEBUG, "config: totaldownload\n");
 		} else if(strcmp(key, "CheckSpace") == 0) {
 			config->checkspace = 1;
+		} else if(strcmp(key, "Color") == 0) {
+			if(config->color == PM_COLOR_UNSET) {
+				config->color = isatty(fileno(stdout)) ? PM_COLOR_ON : PM_COLOR_OFF;
+				enable_colors(config->color);
+			}
 		} else {
 			pm_printf(ALPM_LOG_WARNING,
 					_("config file %s, line %d: directive '%s' in section '%s' not recognized.\n"),
@@ -420,14 +508,32 @@ static int _parse_options(const char *key, char *value,
 			setrepeatingoption(value, "IgnoreGroup", &(config->ignoregrp));
 		} else if(strcmp(key, "HoldPkg") == 0) {
 			setrepeatingoption(value, "HoldPkg", &(config->holdpkg));
-		} else if(strcmp(key, "SyncFirst") == 0) {
-			setrepeatingoption(value, "SyncFirst", &(config->syncfirst));
 		} else if(strcmp(key, "CacheDir") == 0) {
 			setrepeatingoption(value, "CacheDir", &(config->cachedirs));
 		} else if(strcmp(key, "Architecture") == 0) {
 			if(!config->arch) {
 				config_set_arch(value);
 			}
+		} else if(strcmp(key, "UseDelta") == 0) {
+			double ratio;
+			char *endptr;
+			const char *oldlocale;
+
+			/* set the locale to 'C' for consistant decimal parsing (0.7 and never
+			 * 0,7) from config files, then restore old setting when we are done */
+			oldlocale = setlocale(LC_NUMERIC, NULL);
+			setlocale(LC_NUMERIC, "C");
+			ratio = strtod(value, &endptr);
+			setlocale(LC_NUMERIC, oldlocale);
+
+			if(*endptr != '\0' || ratio < 0.0 || ratio > 2.0) {
+				pm_printf(ALPM_LOG_ERROR,
+						_("config file %s, line %d: invalid value for '%s' : '%s'\n"),
+						file, linenum, "UseDelta", value);
+				return 1;
+			}
+			config->deltaratio = ratio;
+			pm_printf(ALPM_LOG_DEBUG, "config: usedelta = %f\n", ratio);
 		} else if(strcmp(key, "DBPath") == 0) {
 			/* don't overwrite a path specified on the command line */
 			if(!config->dbpath) {
@@ -469,6 +575,22 @@ static int _parse_options(const char *key, char *value,
 				return 1;
 			}
 			FREELIST(values);
+		} else if(strcmp(key, "LocalFileSigLevel") == 0) {
+			alpm_list_t *values = NULL;
+			setrepeatingoption(value, "LocalFileSigLevel", &values);
+			if(process_siglevel(values, &config->localfilesiglevel, file, linenum)) {
+				FREELIST(values);
+				return 1;
+			}
+			FREELIST(values);
+		} else if(strcmp(key, "RemoteFileSigLevel") == 0) {
+			alpm_list_t *values = NULL;
+			setrepeatingoption(value, "RemoteFileSigLevel", &values);
+			if(process_siglevel(values, &config->remotefilesiglevel, file, linenum)) {
+				FREELIST(values);
+				return 1;
+			}
+			FREELIST(values);
 		} else {
 			pm_printf(ALPM_LOG_WARNING,
 					_("config file %s, line %d: directive '%s' in section '%s' not recognized.\n"),
@@ -522,7 +644,7 @@ static int _add_mirror(alpm_db_t *db, char *value)
 static int setup_libalpm(void)
 {
 	int ret = 0;
-	enum _alpm_errno_t err;
+	alpm_errno_t err;
 	alpm_handle_t *handle;
 
 	pm_printf(ALPM_LOG_DEBUG, "setup_libalpm called\n");
@@ -591,6 +713,11 @@ static int setup_libalpm(void)
 
 	alpm_option_set_default_siglevel(handle, config->siglevel);
 
+	merge_siglevel(&config->siglevel, &config->localfilesiglevel);
+	merge_siglevel(&config->siglevel, &config->remotefilesiglevel);
+	alpm_option_set_local_file_siglevel(handle, config->localfilesiglevel);
+	alpm_option_set_remote_file_siglevel(handle, config->remotefilesiglevel);
+
 	if(config->xfercommand) {
 		alpm_option_set_fetchcb(handle, download_with_xfercommand);
 	} else if(!(alpm_capabilities() & ALPM_CAPABILITY_DOWNLOADER)) {
@@ -604,7 +731,7 @@ static int setup_libalpm(void)
 	alpm_option_set_arch(handle, config->arch);
 	alpm_option_set_checkspace(handle, config->checkspace);
 	alpm_option_set_usesyslog(handle, config->usesyslog);
-	alpm_option_set_usedelta(handle, config->usedelta);
+	alpm_option_set_deltaratio(handle, config->deltaratio);
 
 	alpm_option_set_ignorepkgs(handle, config->ignorepkg);
 	alpm_option_set_ignoregroups(handle, config->ignoregrp);
@@ -650,7 +777,7 @@ static int finish_section(struct section_t *section, int parse_options)
 	}
 
 	/* if we are not looking at options sections only, register a db */
-	db = alpm_db_register_sync(config->handle, section->name, section->siglevel);
+	db = alpm_register_syncdb(config->handle, section->name, section->siglevel);
 	if(db == NULL) {
 		pm_printf(ALPM_LOG_ERROR, _("could not register '%s' database (%s)\n"),
 				section->name, alpm_strerror(alpm_errno(config->handle)));
@@ -659,7 +786,7 @@ static int finish_section(struct section_t *section, int parse_options)
 	}
 
 	for(i = section->servers; i; i = alpm_list_next(i)) {
-		char *value = alpm_list_getdata(i);
+		char *value = i->data;
 		if(_add_mirror(db, value) != 0) {
 			pm_printf(ALPM_LOG_ERROR,
 					_("could not add mirror '%s' to database '%s' (%s)\n"),
@@ -709,7 +836,8 @@ static int _parseconfig(const char *file, struct section_t *section,
 	pm_printf(ALPM_LOG_DEBUG, "config: attempting to read file %s\n", file);
 	fp = fopen(file, "r");
 	if(fp == NULL) {
-		pm_printf(ALPM_LOG_ERROR, _("config file %s could not be read.\n"), file);
+		pm_printf(ALPM_LOG_ERROR, _("config file %s could not be read: %s\n"),
+				file, strerror(errno));
 		ret = 1;
 		goto cleanup;
 	}
@@ -725,8 +853,7 @@ static int _parseconfig(const char *file, struct section_t *section,
 			*ptr = '\0';
 		}
 
-		strtrim(line);
-		line_len = strlen(line);
+		line_len = strtrim(line);
 
 		if(line_len == 0) {
 			continue;
@@ -822,7 +949,7 @@ static int _parseconfig(const char *file, struct section_t *section,
 			if((ret = _parse_options(key, value, file, linenum)) != 0) {
 				goto cleanup;
 			}
-		} else if (!parse_options && !section->is_options) {
+		} else if(!parse_options && !section->is_options) {
 			/* ... or in a repo section */
 			if(strcmp(key, "Server") == 0) {
 				if(value == NULL) {
@@ -859,7 +986,7 @@ static int _parseconfig(const char *file, struct section_t *section,
 	}
 
 cleanup:
-	if (fp) {
+	if(fp) {
 		fclose(fp);
 	}
 	pm_printf(ALPM_LOG_DEBUG, "config: finished parsing %s\n", file);
