@@ -1,7 +1,7 @@
 /*
  *  deptest.c
  *
- *  Copyright (c) 2006-2011 Pacman Development Team <pacman-dev@archlinux.org>
+ *  Copyright (c) 2006-2013 Pacman Development Team <pacman-dev@archlinux.org>
  *  Copyright (c) 2002-2006 by Judd Vinet <jvinet@zeroflux.org>
  *
  *  This program is free software; you can redistribute it and/or modify
@@ -18,8 +18,6 @@
  *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
  */
 
-#include "config.h"
-
 #include <stdio.h>
 
 #include <alpm.h>
@@ -33,10 +31,10 @@ int pacman_deptest(alpm_list_t *targets)
 {
 	alpm_list_t *i;
 	alpm_list_t *deps = NULL;
-	alpm_db_t *localdb = alpm_option_get_localdb(config->handle);
+	alpm_db_t *localdb = alpm_get_localdb(config->handle);
 
 	for(i = targets; i; i = alpm_list_next(i)) {
-		char *target = alpm_list_getdata(i);
+		char *target = i->data;
 
 		if(!alpm_find_satisfier(alpm_db_get_pkgcache(localdb), target)) {
 			deps = alpm_list_add(deps, target);
@@ -48,7 +46,7 @@ int pacman_deptest(alpm_list_t *targets)
 	}
 
 	for(i = deps; i; i = alpm_list_next(i)) {
-		const char *dep = alpm_list_getdata(i);
+		const char *dep = i->data;
 
 		printf("%s\n", dep);
 	}
