@@ -1,7 +1,7 @@
 /*
  *  package.h
  *
- *  Copyright (c) 2006-2011 Pacman Development Team <pacman-dev@archlinux.org>
+ *  Copyright (c) 2006-2013 Pacman Development Team <pacman-dev@archlinux.org>
  *  Copyright (c) 2002-2006 by Judd Vinet <jvinet@zeroflux.org>
  *
  *  This program is free software; you can redistribute it and/or modify
@@ -28,6 +28,9 @@ void dump_pkg_backups(alpm_pkg_t *pkg);
 void dump_pkg_files(alpm_pkg_t *pkg, int quiet);
 void dump_pkg_changelog(alpm_pkg_t *pkg);
 
+void print_installed(alpm_db_t *db_local, alpm_pkg_t *pkg);
+int dump_pkg_search(alpm_db_t *db, alpm_list_t *targets, int show_status);
+
 #endif /* _PM_PACKAGE_H */
 
 /* vim: set ts=2 sw=2 noet: */
