@@ -1,7 +1,7 @@
 /*
  *  pacman.c
  *
- *  Copyright (c) 2006-2011 Pacman Development Team <pacman-dev@archlinux.org>
+ *  Copyright (c) 2006-2013 Pacman Development Team <pacman-dev@archlinux.org>
  *  Copyright (c) 2002-2006 by Judd Vinet <jvinet@zeroflux.org>
  *
  *  This program is free software; you can redistribute it and/or modify
@@ -18,8 +18,6 @@
  *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
  */
 
-#include "config.h"
-
 /* special handling of package version for GIT */
 #if defined(GIT_VERSION)
 #undef PACKAGE_VERSION
@@ -39,9 +37,6 @@
 #include <sys/utsname.h> /* uname */
 #include <locale.h> /* setlocale */
 #include <errno.h>
-#if defined(PACMAN_DEBUG) && defined(HAVE_MCHECK_H)
-#include <mcheck.h> /* debug tracing (mtrace) */
-#endif
 
 /* alpm */
 #include <alpm.h>
@@ -147,6 +142,7 @@ static void usage(int op, const char * const myname)
 			addlist(_("  -k, --check          check that the files owned by the package(s) are present\n"));
 			addlist(_("  -l, --list           list the contents of the queried package\n"));
 			addlist(_("  -m, --foreign        list installed packages not found in sync db(s) [filter]\n"));
+			addlist(_("  -n, --native         list installed packages only found in sync db(s) [filter]\n"));
 			addlist(_("  -o, --owns <file>    query the package that owns <file>\n"));
 			addlist(_("  -p, --file <package> query a package file instead of the database\n"));
 			addlist(_("  -q, --quiet          show less information for query and search\n"));
@@ -178,7 +174,7 @@ static void usage(int op, const char * const myname)
 		switch(op) {
 			case PM_OP_SYNC:
 			case PM_OP_UPGRADE:
-				addlist(_("  -f, --force          force install, overwrite conflicting files\n"));
+				addlist(_("      --force          force install, overwrite conflicting files\n"));
 				addlist(_("      --asdeps         install packages as non-explicitly installed\n"));
 				addlist(_("      --asexplicit     install packages as explicitly installed\n"));
 				addlist(_("      --ignore <pkg>   ignore a package upgrade (can be used more than once)\n"));
@@ -201,6 +197,7 @@ static void usage(int op, const char * const myname)
 		addlist(_("  -v, --verbose        be verbose\n"));
 		addlist(_("      --arch <arch>    set an alternate architecture\n"));
 		addlist(_("      --cachedir <dir> set an alternate package cache location\n"));
+		addlist(_("      --color <when>   colorize the output\n"));
 		addlist(_("      --config <path>  set an alternate configuration file\n"));
 		addlist(_("      --debug          display debug messages\n"));
 		addlist(_("      --gpgdir <path>  set an alternate home directory for GnuPG\n"));
@@ -208,8 +205,8 @@ static void usage(int op, const char * const myname)
 		addlist(_("      --noconfirm      do not ask for any confirmation\n"));
 	}
 	list = alpm_list_msort(list, alpm_list_count(list), options_cmp);
-	for (i = list; i; i = alpm_list_next(i)) {
-		printf("%s", (char *)alpm_list_getdata(i));
+	for(i = list; i; i = alpm_list_next(i)) {
+		fputs((const char *)i->data, stdout);
 	}
 	alpm_list_free(list);
 #undef addlist
@@ -221,7 +218,7 @@ static void version(void)
 {
 	printf("\n");
 	printf(" .--.                  Pacman v%s - libalpm v%s\n", PACKAGE_VERSION, alpm_version());
-	printf("/ _.-' .-.  .-.  .-.   Copyright (C) 2006-2012 Pacman Development Team\n");
+	printf("/ _.-' .-.  .-.  .-.   Copyright (C) 2006-2013 Pacman Development Team\n");
 	printf("\\  '-. '-'  '-'  '-'   Copyright (C) 2002-2006 Judd Vinet\n");
 	printf(" '--'\n");
 	printf(_("                       This program may be freely redistributed under\n"
@@ -262,7 +259,8 @@ static void setuseragent(void)
  *
  * @param ret the return value
  */
-static void cleanup(int ret) {
+static void cleanup(int ret)
+{
 	/* free alpm library resources */
 	if(config->handle && alpm_release(config->handle) == -1) {
 		pm_printf(ALPM_LOG_ERROR, "error releasing alpm library\n");
@@ -397,6 +395,20 @@ static int parsearg_global(int opt)
 			check_optarg();
 			config->cachedirs = alpm_list_add(config->cachedirs, strdup(optarg));
 			break;
+		case OP_COLOR:
+			if(strcmp("never", optarg) == 0) {
+				config->color = PM_COLOR_OFF;
+			} else if(strcmp("auto", optarg) == 0) {
+				config->color = isatty(fileno(stdout)) ? PM_COLOR_ON : PM_COLOR_OFF;
+			} else if(strcmp("always", optarg) == 0) {
+				config->color = PM_COLOR_ON;
+			} else {
+				pm_printf(ALPM_LOG_ERROR, _("invalid argument '%s' for %s\n"),
+						optarg, "--color");
+				return 1;
+			}
+			enable_colors(config->color);
+			break;
 		case OP_CONFIG:
 			check_optarg();
 			if(config->configfile) {
@@ -464,9 +476,10 @@ static int parsearg_query(int opt)
 		case 'e': config->op_q_explicit = 1; break;
 		case 'g': (config->group)++; break;
 		case 'i': (config->op_q_info)++; break;
-		case 'k': config->op_q_check = 1; break;
+		case 'k': (config->op_q_check)++; break;
 		case 'l': config->op_q_list = 1; break;
-		case 'm': config->op_q_foreign = 1; break;
+		case 'm': config->op_q_locality |= PKG_LOCALITY_LOCAL; break;
+		case 'n': config->op_q_locality |= PKG_LOCALITY_FOREIGN; break;
 		case 'o': config->op_q_owns = 1; break;
 		case 'p': config->op_q_isfile = 1; break;
 		case 'q': config->quiet = 1; break;
@@ -528,7 +541,7 @@ static int parsearg_upgrade(int opt)
 	if(parsearg_trans(opt) == 0)
 		return 0;
 	switch(opt) {
-		case 'f': config->flags |= ALPM_TRANS_FLAG_FORCE; break;
+		case OP_FORCE: config->flags |= ALPM_TRANS_FLAG_FORCE; break;
 		case OP_ASDEPS: config->flags |= ALPM_TRANS_FLAG_ALLDEPS; break;
 		case OP_ASEXPLICIT: config->flags |= ALPM_TRANS_FLAG_ALLEXPLICIT; break;
 		case OP_NEEDED: config->flags |= ALPM_TRANS_FLAG_NEEDED; break;
@@ -593,13 +606,13 @@ static int parseargs(int argc, char *argv[])
 		{"nodeps",     no_argument,       0, 'd'},
 		{"deps",       no_argument,       0, 'd'},
 		{"explicit",   no_argument,       0, 'e'},
-		{"force",      no_argument,       0, 'f'},
 		{"groups",     no_argument,       0, 'g'},
 		{"help",       no_argument,       0, 'h'},
 		{"info",       no_argument,       0, 'i'},
 		{"check",      no_argument,       0, 'k'},
 		{"list",       no_argument,       0, 'l'},
 		{"foreign",    no_argument,       0, 'm'},
+		{"native",     no_argument,       0, 'n'},
 		{"nosave",     no_argument,       0, 'n'},
 		{"owns",       no_argument,       0, 'o'},
 		{"file",       no_argument,       0, 'p'},
@@ -620,6 +633,7 @@ static int parseargs(int argc, char *argv[])
 		{"config",     required_argument, 0, OP_CONFIG},
 		{"ignore",     required_argument, 0, OP_IGNORE},
 		{"debug",      optional_argument, 0, OP_DEBUG},
+		{"force",      no_argument,       0, OP_FORCE},
 		{"noprogressbar", no_argument,    0, OP_NOPROGRESSBAR},
 		{"noscriptlet", no_argument,      0, OP_NOSCRIPTLET},
 		{"ask",        required_argument, 0, OP_ASK},
@@ -633,6 +647,7 @@ static int parseargs(int argc, char *argv[])
 		{"print-format", required_argument, 0, OP_PRINTFORMAT},
 		{"gpgdir",     required_argument, 0, OP_GPGDIR},
 		{"dbonly",     no_argument,       0, OP_DBONLY},
+		{"color",      required_argument, 0, OP_COLOR},
 		{0, 0, 0, 0}
 	};
 
@@ -724,7 +739,7 @@ static int parseargs(int argc, char *argv[])
  * @param argc
  * @param argv
  */
-static void cl_to_log(int argc, char* argv[])
+static void cl_to_log(int argc, char *argv[])
 {
 	size_t size = 0;
 	int i;
@@ -745,7 +760,8 @@ static void cl_to_log(int argc, char* argv[])
 		*p++ = ' ';
 	}
 	strcpy(p, argv[i]);
-	alpm_logaction(config->handle, "Running '%s'\n", cl_text);
+	alpm_logaction(config->handle, PACMAN_CALLER_PREFIX,
+			"Running '%s'\n", cl_text);
 	free(cl_text);
 }
 
@@ -760,15 +776,7 @@ int main(int argc, char *argv[])
 	size_t i;
 	struct sigaction new_action, old_action;
 	const int signals[] = { SIGHUP, SIGINT, SIGTERM, SIGSEGV };
-#if defined(HAVE_GETEUID) && !defined(CYGWIN)
-	/* geteuid undefined in CYGWIN */
 	uid_t myuid = geteuid();
-#endif
-
-#if defined(PACMAN_DEBUG) && defined(HAVE_MCHECK_H)
-	/*setenv("MALLOC_TRACE","pacman.mtrace", 0);*/
-	mtrace();
-#endif
 
 	/* Set signal handlers */
 	/* Set up the structure to specify the new action. */
@@ -817,55 +825,66 @@ int main(int argc, char *argv[])
 	}
 
 	/* we support reading targets from stdin if a cmdline parameter is '-' */
-	if(!isatty(fileno(stdin)) && alpm_list_find_str(pm_targets, "-")) {
-		size_t current_size = PATH_MAX, i = 0;
-		char *line = malloc(current_size);
+	if(alpm_list_find_str(pm_targets, "-")) {
+		if(!isatty(fileno(stdin))) {
+			size_t current_size = PATH_MAX;
+			char *vdata, *line = malloc(current_size);
 
-		/* remove the '-' from the list */
-		pm_targets = alpm_list_remove_str(pm_targets, "-", NULL);
+			/* remove the '-' from the list */
+			pm_targets = alpm_list_remove_str(pm_targets, "-", &vdata);
+			free(vdata);
 
-		while((line[i] = (char)fgetc(stdin)) != EOF) {
-			if(isspace((unsigned char)line[i])) {
-				/* avoid adding zero length arg when multiple spaces separate args */
-				if(i > 0) {
-					line[i] = '\0';
-					pm_targets = alpm_list_add(pm_targets, strdup(line));
-					i = 0;
-				}
-			} else {
-				i++;
-				/* we may be at the end of our allocated buffer now */
-				if(i >= current_size) {
-					char *new = realloc(line, current_size * 2);
-					if(new) {
-						line = new;
-						current_size *= 2;
-					} else {
-						free(line);
-						line = NULL;
-						break;
+			i = 0;
+			while((line[i] = (char)fgetc(stdin)) != EOF) {
+				if(isspace((unsigned char)line[i])) {
+					/* avoid adding zero length arg when multiple spaces separate args */
+					if(i > 0) {
+						line[i] = '\0';
+						pm_targets = alpm_list_add(pm_targets, strdup(line));
+						i = 0;
+					}
+				} else {
+					i++;
+					/* we may be at the end of our allocated buffer now */
+					if(i >= current_size) {
+						char *new = realloc(line, current_size * 2);
+						if(new) {
+							line = new;
+							current_size *= 2;
+						} else {
+							free(line);
+							line = NULL;
+							break;
+						}
 					}
 				}
 			}
-		}
-		/* check for memory exhaustion */
-		if(!line) {
-			pm_printf(ALPM_LOG_ERROR, _("memory exhausted in argument parsing\n"));
-			cleanup(EXIT_FAILURE);
-		}
 
-		/* end of stream -- check for data still in line buffer */
-		if(i > 0) {
-			line[i] = '\0';
-			pm_targets = alpm_list_add(pm_targets, strdup(line));
-		}
-		free(line);
-		if(!freopen(ctermid(NULL), "r", stdin)) {
-			pm_printf(ALPM_LOG_ERROR, _("failed to reopen stdin for reading: (%s)\n"),
-					strerror(errno));
+			/* check for memory exhaustion */
+			if(!line) {
+				pm_printf(ALPM_LOG_ERROR, _("memory exhausted in argument parsing\n"));
+				cleanup(EXIT_FAILURE);
+			}
+
+			/* end of stream -- check for data still in line buffer */
+			if(i > 0) {
+				line[i] = '\0';
+				pm_targets = alpm_list_add(pm_targets, strdup(line));
+			}
+			free(line);
+			if(!freopen(ctermid(NULL), "r", stdin)) {
+				pm_printf(ALPM_LOG_ERROR, _("failed to reopen stdin for reading: (%s)\n"),
+						strerror(errno));
+			}
+		} else {
+			/* do not read stdin from terminal */
+			pm_printf(ALPM_LOG_ERROR, _("argument '-' specified without input on stdin\n"));
+			cleanup(1);
 		}
 	}
 
+	pm_printf(ALPM_LOG_DEBUG, "pacman v%s - libalpm v%s\n", PACKAGE_VERSION, alpm_version());
+
 	/* parse the config file */
 	ret = parseconfig(config->configfile);
 	if(ret != 0) {
@@ -886,28 +905,26 @@ int main(int argc, char *argv[])
 		config->logmask &= ~ALPM_LOG_WARNING;
 	}
 
-#if defined(HAVE_GETEUID) && !defined(CYGWIN)
 	/* check if we have sufficient permission for the requested operation */
 	if(myuid > 0 && needs_root()) {
 		pm_printf(ALPM_LOG_ERROR, _("you cannot perform this operation unless you are root.\n"));
 		cleanup(EXIT_FAILURE);
 	}
-#endif
 
 	if(config->verbose > 0) {
-		alpm_list_t *i;
+		alpm_list_t *j;
 		printf("Root      : %s\n", alpm_option_get_root(config->handle));
 		printf("Conf File : %s\n", config->configfile);
 		printf("DB Path   : %s\n", alpm_option_get_dbpath(config->handle));
 		printf("Cache Dirs: ");
-		for(i = alpm_option_get_cachedirs(config->handle); i; i = alpm_list_next(i)) {
-			printf("%s  ", (char *)alpm_list_getdata(i));
+		for(j = alpm_option_get_cachedirs(config->handle); j; j = alpm_list_next(j)) {
+			printf("%s  ", (const char *)j->data);
 		}
 		printf("\n");
 		printf("Lock File : %s\n", alpm_option_get_lockfile(config->handle));
 		printf("Log File  : %s\n", alpm_option_get_logfile(config->handle));
 		printf("GPG Dir   : %s\n", alpm_option_get_gpgdir(config->handle));
-		list_display("Targets   :", pm_targets);
+		list_display("Targets   :", pm_targets, 0);
 	}
 
 	/* Log command line */
