@@ -24,6 +24,7 @@
 
 #define BASENAME "testpkg"
 
+__attribute__((format(printf, 2, 0)))
 static void output_cb(alpm_loglevel_t level, const char *fmt, va_list args)
 {
 	if(fmt[0] == '\0') {
@@ -41,7 +42,7 @@ int main(int argc, char *argv[])
 {
 	int retval = 1; /* default = false */
 	alpm_handle_t *handle;
-	enum _alpm_errno_t err;
+	alpm_errno_t err;
 	alpm_pkg_t *pkg = NULL;
 	const alpm_siglevel_t level = ALPM_SIG_PACKAGE | ALPM_SIG_PACKAGE_OPTIONAL;
 
@@ -59,6 +60,9 @@ int main(int argc, char *argv[])
 	/* let us get log messages from libalpm */
 	alpm_option_set_logcb(handle, output_cb);
 
+	/* set gpgdir to default */
+	alpm_option_set_gpgdir(handle, GPGDIR);
+
 	if(alpm_pkg_load(handle, argv[1], 1, level, &pkg) == -1
 			|| pkg == NULL) {
 		err = alpm_errno(handle);
