@@ -1,6 +1,7 @@
-#! /usr/bin/python
+#! /usr/bin/python2
 #
 #  Copyright (c) 2006 by Aurelien Foret <orelien@chez.com>
+#  Copyright (c) 2006-2013 Pacman Development Team <pacman-dev@archlinux.org>
 #
 #  This program is free software; you can redistribute it and/or modify
 #  it under the terms of the GNU General Public License as published by
@@ -87,8 +88,11 @@ class pmrule(object):
                     if not value in newpkg.depends:
                         success = 0
                 elif case == "OPTDEPENDS":
-                    if not value in newpkg.optdepends:
-                        success = 0
+                    success = 0
+                    for optdep in newpkg.optdepends:
+                        if value == optdep.split(':', 1)[0]:
+                            success = 1
+                            break
                 elif case == "REASON":
                     if newpkg.reason != int(value):
                         success = 0
