@@ -1,6 +1,6 @@
 self.description = "Synchronize the local database"
 
-if len("-lcurl") == 0:
+if len("") == 0:
 	self.option['XferCommand'] = ['/usr/bin/curl %u > %o']
 
 sp1 = pmpkg("spkg1", "1.0-1")
