@@ -20,7 +20,7 @@ p2.files = ["bin/foobar"]
 for p in p1, p2:
 	self.addpkg(p)
 
-self.args = "-Uf %s" % " ".join([p.filename() for p in p1, p2])
+self.args = "-U --force %s" % " ".join([p.filename() for p in p1, p2])
 
 self.addrule("PACMAN_RETCODE=0")
 for p in p1, p2:
@@ -29,5 +29,3 @@ self.addrule("FILE_MODIFIED=bin/dummy")
 self.addrule("FILE_MODIFIED=bin/foobar")
 self.addrule("FILE_EXIST=usr/share/file")
 self.addrule("FILE_MODIFIED=usr/share/file")
-
-self.expectfailure = True
